# -*- coding: utf-8 -*-
"""
Test the RescaleToBound class.
"""
import numpy as np
import pytest
from unittest.mock import MagicMock, call, create_autospec, patch

from nessai.reparameterisations import RescaleToBounds
from nessai.livepoint import (
    get_dtype,
    empty_structured_array,
    numpy_array_to_live_points,
)
from nessai.utils.testing import assert_structured_arrays_equal

# Tolerances for assert_allclose
atol = 1e-15
rtol = 1e-15


@pytest.fixture
def reparam():
    return create_autospec(RescaleToBounds)


@pytest.fixture()
def reparameterisation(model):
    def _get_reparameterisation(kwargs):
        return RescaleToBounds(
            parameters=model.names, prior_bounds=model.bounds, **kwargs
        )

    return _get_reparameterisation


@pytest.fixture(scope="function")
def is_invertible(model, n=100):
    def test_invertibility(reparam, model=model, decimal=16):
        x = model.new_point(N=n)
        x_prime = empty_structured_array(n, names=reparam.prime_parameters)
        log_j = np.zeros(n)
        assert x.size == x_prime.size

        x_re, x_prime_re, log_j_re = reparam.reparameterise(x, x_prime, log_j)

        x_in = empty_structured_array(x_re.size, reparam.parameters)
        log_j = np.zeros(x_re.size)

        x_inv, x_prime_inv, log_j_inv = reparam.inverse_reparameterise(
            x_in, x_prime_re, log_j
        )

        m = x_re.size // n
        for i in range(m):
            start, end = (i * n), (i + 1) * n
            for name in x.dtype.names:
                np.testing.assert_array_almost_equal(
                    x[name],
                    x_re[name][start:end],
                    decimal=decimal,
                )
                np.testing.assert_array_almost_equal(
                    x[name],
                    x_inv[name][start:end],
                    decimal=decimal,
                )
            for name in x_prime.dtype.names:
                np.testing.assert_array_almost_equal(
                    x_prime_re[name],
                    x_prime_inv[name],
                    decimal=decimal,
                )
            np.testing.assert_array_almost_equal(
                log_j_re, -log_j_inv, decimal=decimal
            )

        return True

    return test_invertibility


@pytest.mark.parametrize(
    "input, expected_value",
    [
        (None, {"x": [-1, 1], "y": [-1, 1]}),
        ([0, 1], {"x": [0, 1], "y": [0, 1]}),
        ({"x": [0, 1], "y": [-1, 1]}, {"x": [0, 1], "y": [-1, 1]}),
    ],
)
def test_rescale_bounds_config(reparam, input, expected_value):
    """Assert the rescale bounds are set correctly."""
    RescaleToBounds.__init__(
        reparam,
        parameters=["x", "y"],
        prior_bounds={"x": [-1, 1], "y": [0, 1]},
        rescale_bounds=input,
    )
    assert reparam.rescale_bounds == expected_value


def test_rescale_bounds_dict_missing_params(reparam):
    """Assert an error is raised if the rescale_bounds dict is missing a
    parameter.
    """
    with pytest.raises(RuntimeError) as excinfo:
        RescaleToBounds.__init__(
            reparam,
            parameters=["x", "y"],
            prior_bounds={"x": [-1, 1], "y": [0, 1]},
            rescale_bounds={"x": [0, 1]},
        )
    assert "Missing rescale bounds for parameters" in str(excinfo.value)


def test_rescale_bounds_incorrect_type(reparam):
    """Assert an error is raised if the rescale_bounds is an invalid type."""
    with pytest.raises(TypeError) as excinfo:
        RescaleToBounds.__init__(
            reparam,
            parameters=["x", "y"],
            prior_bounds={"x": [-1, 1], "y": [0, 1]},
            rescale_bounds=1,
        )
    assert "must be an instance of list or dict" in str(excinfo.value)


@pytest.mark.parametrize(
    "input, expected_value",
    [
        (True, {"x": "split", "y": "split"}),
        (False, False),
        (["x"], {"x": "split"}),
        ({"x": "split"}, {"x": "split"}),
        (None, False),
    ],
)
def test_boundary_inversion_config(reparam, input, expected_value):
    """Assert the boundary inversion dict is set correctly"""
    RescaleToBounds.__init__(
        reparam,
        parameters=["x", "y"],
        prior_bounds={"x": [0, 1], "y": [0, 1]},
        boundary_inversion=input,
    )
    assert reparam.boundary_inversion == expected_value


def test_boundary_inversion_invalid_type(reparam):
    """Assert an error is raised in the type is invalid"""
    with pytest.raises(TypeError) as excinfo:
        RescaleToBounds.__init__(
            reparam,
            parameters="x",
            prior_bounds=[0, 1],
            boundary_inversion="Yes",
        )
    assert "boundary_inversion must be a list, dict or bool" in str(
        excinfo.value
    )


def test_detect_edges_without_inversion(reparam):
    """Assert detect edges cannot be used with boundary inversion"""
    with pytest.raises(RuntimeError) as excinfo:
        RescaleToBounds.__init__(
            reparam,
            parameters=["x", "y"],
            prior_bounds={"x": [-1, 1], "y": [0, 1]},
            detect_edges=True,
        )
    assert "Must enable boundary inversion to use detect edges" in str(
        excinfo.value
    )


def test_set_bounds(reparam):
    """Test the set bounds method."""
    reparam.parameters = ["x"]
    reparam.rescale_bounds = {"x": np.array([-1, 1])}
    reparam.pre_rescaling = lambda x: (x / 2, np.zeros_like(x))
    reparam.offsets = {"x": 1}
    RescaleToBounds.set_bounds(reparam, {"x": np.array([-10, 10])})
    np.testing.assert_array_equal(reparam.pre_prior_bounds["x"], [-5, 5])
    np.testing.assert_array_equal(reparam.bounds["x"], [-6, 4])


def test_set_offets(reparam):
    """Assert the offset are set correctly"""
    reparam.pre_rescaling = lambda x: (x / 2, 0.0)

    RescaleToBounds.__init__(
        reparam,
        parameters=["x", "y"],
        prior_bounds={"x": [8, 32], "y": [2, 4]},
        offset=True,
    )

    assert reparam.offsets == {"x": 10.0, "y": 1.5}


def test_reset_inversion(reparam):
    """Assert the edges are reset correctly"""
    reparam.parameters = ["x", "y"]
    reparam._edges = {"x": [-10, 10], "y": [-5, 5]}
    RescaleToBounds.reset_inversion(reparam)
    assert reparam._edges == {"x": None, "y": None}


def test_reset_inversion_no_edges(reparam):
    """Assert the edges are reset not reset if _edges is None"""
    reparam.parameters = ["x", "y"]
    reparam._edges = None
    RescaleToBounds.reset_inversion(reparam)
    assert reparam._edges is None


def test_update(reparam):
    """Assert update calls the correct methods"""
    reparam.reset_inversion = MagicMock()
    reparam.update_bounds = MagicMock()
    x = np.array((1, 2), dtype=[("x", "f8"), ("y", "f8")])
    RescaleToBounds.update(reparam, x)
    reparam.reset_inversion.assert_called_once()
    reparam.update_bounds.assert_called_once_with(x)


def test_x_prime_log_prior_error(reparam):
    """Assert an error is raised if the prime prior is not defined."""
    reparam.has_prime_prior = False
    with pytest.raises(RuntimeError) as excinfo:
        RescaleToBounds.x_prime_log_prior(reparam, 0.1)
    assert "Prime prior is not configured" in str(excinfo.value)


def test_default_pre_rescaling(reparam):
    """Assert the default pre-rescaling is the identity"""
    x = np.array([1, 2, 3])
    expected_log_j = np.zeros(3)
    x_out, log_j = RescaleToBounds.pre_rescaling(reparam, x)
    x_out_inv, log_j_inv = RescaleToBounds.pre_rescaling_inv(reparam, x)

    np.testing.assert_array_equal(x_out, x)
    np.testing.assert_array_equal(x_out_inv, x)
    np.testing.assert_array_equal(log_j, expected_log_j)
    np.testing.assert_array_equal(log_j_inv, expected_log_j)


def test_default_post_rescaling(reparam):
    """Assert the default post-rescaling is the identity"""
    x = np.array([1, 2, 3])
    expected_log_j = np.zeros(3)
    x_out, log_j = RescaleToBounds.post_rescaling(reparam, x)
    x_out_inv, log_j_inv = RescaleToBounds.post_rescaling_inv(reparam, x)

    np.testing.assert_array_equal(x_out, x)
    np.testing.assert_array_equal(x_out_inv, x)
    np.testing.assert_array_equal(log_j, expected_log_j)
    np.testing.assert_array_equal(log_j_inv, expected_log_j)


def test_configure_pre_rescaling_none(reparam):
    """Test the configuration of the pre-rescaling if it is None"""
    RescaleToBounds.configure_pre_rescaling(reparam, None)
    assert reparam.has_pre_rescaling is False


def test_configure_post_rescaling_none(reparam):
    """Test the configuration of the post-rescaling if it is None"""
    RescaleToBounds.configure_post_rescaling(reparam, None)
    assert reparam.has_post_rescaling is False


def test_pre_rescaling_with_functions(reparam):
    """Assert that specifying functions works as intended"""
    rescaling = (np.exp, np.log)
    RescaleToBounds.configure_pre_rescaling(reparam, rescaling)
    assert reparam.has_pre_rescaling is True
    assert reparam.pre_rescaling is np.exp
    assert reparam.pre_rescaling_inv is np.log


def test_post_rescaling_with_functions(reparam):
    """Assert that specifying functions works as intended"""
    rescaling = (np.exp, np.log)
    RescaleToBounds.configure_post_rescaling(reparam, rescaling)
    assert reparam.has_post_rescaling is True
    assert reparam.has_prime_prior is False
    assert reparam.post_rescaling is np.exp
    assert reparam.post_rescaling_inv is np.log


def test_pre_rescaling_with_str(reparam):
    """Assert that specifying a str works as intended"""
    from nessai.utils.rescaling import rescaling_functions

    rescaling = "logit"
    RescaleToBounds.configure_pre_rescaling(reparam, rescaling)
    assert reparam.has_pre_rescaling is True
    assert reparam.pre_rescaling is rescaling_functions["logit"][0]
    assert reparam.pre_rescaling_inv is rescaling_functions["logit"][1]


def test_pre_rescaling_with_invalid_str(reparam):
    """Assert an error is raised if the rescaling is not recognised"""
    rescaling = "not_a_rescaling"
    with pytest.raises(RuntimeError) as excinfo:
        RescaleToBounds.configure_pre_rescaling(reparam, rescaling)
    assert "Unknown rescaling function: not_a_rescaling" in str(excinfo.value)


def test_post_rescaling_with_invalid_str(reparam):
    """Assert an error is raised if the rescaling is not recognised"""
    rescaling = "not_a_rescaling"
    with pytest.raises(RuntimeError) as excinfo:
        RescaleToBounds.configure_post_rescaling(reparam, rescaling)
    assert "Unknown rescaling function: not_a_rescaling" in str(excinfo.value)


@pytest.mark.parametrize("rescaling", ["log", "logit"])
def test_post_rescaling_with_str(reparam, rescaling):
    """Assert that specifying a str works as intended.

    Also test the config for the logit
    """
    reparam._update_bounds = False
    reparam.parameters = ["x"]
    from nessai.utils.rescaling import rescaling_functions

    RescaleToBounds.configure_post_rescaling(reparam, rescaling)
    assert reparam.has_post_rescaling is True
    assert reparam.has_prime_prior is False
    assert reparam.post_rescaling is rescaling_functions[rescaling][0]
    assert reparam.post_rescaling_inv is rescaling_functions[rescaling][1]
    assert reparam.rescale_bounds == {"x": [0, 1]}


@pytest.mark.parametrize("rescaling", ["log", "logit"])
def test_post_rescaling_with_logit_update_bounds(reparam, rescaling):
    """Assert an error is raised if using logit and update bounds"""
    reparam._update_bounds = True
    with pytest.raises(
        RuntimeError, match=r"Cannot use log or logit with update bounds"
    ):
        RescaleToBounds.configure_post_rescaling(reparam, rescaling)


def test_pre_rescaling_invalid_input(reparam):
    """Assert an error is raised if the input isn't a str or tuple"""
    with pytest.raises(RuntimeError) as excinfo:
        RescaleToBounds.configure_pre_rescaling(reparam, (np.exp,))
    assert "Pre-rescaling must be a str or tuple" in str(excinfo.value)


def test_post_rescaling_invalid_input(reparam):
    """Assert an error is raised if the input isn't a str or tuple"""
    with pytest.raises(RuntimeError) as excinfo:
        RescaleToBounds.configure_post_rescaling(reparam, (np.exp,))
    assert "Post-rescaling must be a str or tuple" in str(excinfo.value)


def test_update_bounds_disabled(reparam, caplog):
    """Assert nothing happens in _update_bounds is False"""
    caplog.set_level("DEBUG")
    reparam._update_bounds = False
    RescaleToBounds.update_bounds(reparam, [0, 1])
    assert "Update bounds not enabled" in str(caplog.text)


def test_update_bounds(reparam):
    """Assert the correct values are returned"""
    reparam.offsets = {"x": 0.0, "y": 1.0}
    reparam.pre_rescaling = MagicMock(
        side_effect=lambda x: (x, np.zeros_like(x))
    )
    reparam.parameters = ["x", "y"]
    x = {"x": [-1, 0, 1], "y": [-2, 0, 2]}
    RescaleToBounds.update_bounds(reparam, x)
    reparam.update_prime_prior_bounds.assert_called_once()
    reparam.pre_rescaling.assert_has_calls(
        [call(-1), call(1), call(-2), call(2)]
    )
    assert reparam.bounds == {"x": [-1, 1], "y": [-3, 1]}


def test_reparameterise(reparam):
    """Test the reparameterise function"""
    reparam.has_pre_rescaling = False
    reparam.has_post_rescaling = False
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.offsets = {"x": 1.0}
    reparam.boundary_inversion = False
    x = numpy_array_to_live_points(
        np.array([(1.0,), (2.0,)]), reparam.parameters
    )
    x_prime_in = np.zeros(
        [
            2,
        ],
        dtype=get_dtype(reparam.prime_parameters),
    )
    x_prime_val = np.array([0.0, 0.5])
    log_j = np.zeros(x.size)

    reparam._rescale_to_bounds = MagicMock(
        return_value=(x_prime_val, np.array([0, 0.5]))
    )

    x_out, x_prime_out, log_j_out = RescaleToBounds.reparameterise(
        reparam, x, x_prime_in, log_j
    )

    np.testing.assert_array_equal(
        np.array([0.0, 1.0]),
        reparam._rescale_to_bounds.call_args_list[0][0][0],
    )

    assert reparam._rescale_to_bounds.call_args_list[0][0][1] == "x"

    assert_structured_arrays_equal(x, x_out)
    np.testing.assert_array_equal(x_prime_out["x_prime"], x_prime_val)
    np.testing.assert_array_equal(log_j_out, np.array([0.0, 0.5]))


def test_inverse_reparameterise(reparam):
    """Test the inverse_reparameterise function"""
    reparam.has_pre_rescaling = False
    reparam.has_post_rescaling = False
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.offsets = {"x": 1.0}
    reparam.boundary_inversion = False
    x_prime = numpy_array_to_live_points(
        np.array([(1.0,), (2.0,)]), reparam.prime_parameters
    )
    x_in = np.zeros(
        [
            2,
        ],
        dtype=get_dtype(reparam.parameters),
    )
    x_val = np.array([0.0, 0.5])
    log_j = np.zeros(x_prime.size)

    reparam._inverse_rescale_to_bounds = MagicMock(
        return_value=(x_val, np.array([0, 0.5]))
    )

    x_out, x_prime_out, log_j_out = RescaleToBounds.inverse_reparameterise(
        reparam, x_in, x_prime, log_j
    )

    # x[p] is updated in place, can't test inputs
    reparam._inverse_rescale_to_bounds.assert_called_once()
    assert reparam._inverse_rescale_to_bounds.call_args_list[0][0][1] == "x"

    assert_structured_arrays_equal(x_prime_out, x_prime)
    np.testing.assert_array_equal(x_out["x"], x_val + 1.0)
    np.testing.assert_array_equal(log_j_out, np.array([0.0, 0.5]))


def test_reparameterise_boundary_inversion(reparam):
    """Test the reparameterise function with boundary inversion"""
    reparam.has_pre_rescaling = False
    reparam.has_post_rescaling = False
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.offsets = {"x": 1.0}
    reparam.boundary_inversion = {"x": "split"}
    x = numpy_array_to_live_points(
        np.array([(1.0,), (2.0,)]), reparam.parameters
    )
    inversion_out = numpy_array_to_live_points(
        np.array([(-1.0,), (-2.0,), (1.0,), (2.0,)]), reparam.prime_parameters
    )
    x_prime_in = np.zeros(
        [
            2,
        ],
        dtype=get_dtype(reparam.prime_parameters),
    )
    log_j = np.zeros(x.size)

    x_ex = np.concatenate([x, x])
    x_prime_ex = inversion_out
    log_j_ex = np.array([0, 0.5, 0, 0.5])

    reparam._apply_inversion = MagicMock(
        return_value=(x_ex, x_prime_ex, log_j_ex)
    )

    x_out, x_prime_out, log_j_out = RescaleToBounds.reparameterise(
        reparam,
        x,
        x_prime_in,
        log_j,
        compute_radius=True,
        test="test",
    )

    assert_structured_arrays_equal(
        reparam._apply_inversion.call_args_list[0][0][0], x
    )
    assert_structured_arrays_equal(
        reparam._apply_inversion.call_args_list[0][0][1],
        x_prime_in,
    )
    np.testing.assert_array_equal(
        reparam._apply_inversion.call_args_list[0][0][2], log_j
    )
    assert reparam._apply_inversion.call_args_list[0][0][3] == "x"
    assert reparam._apply_inversion.call_args_list[0][0][4] == "x_prime"
    assert reparam._apply_inversion.call_args_list[0][0][5] is True
    assert reparam._apply_inversion.call_args_list[0][1] == {"test": "test"}

    assert_structured_arrays_equal(x_out, x_ex)
    assert_structured_arrays_equal(x_prime_out, x_prime_ex)
    np.testing.assert_array_equal(log_j_out, log_j_ex)


def test_inverse_reparameterise_boundary_inversion(reparam):
    """Test the inverse_reparameterise function with boundary inversion"""
    reparam.has_pre_rescaling = False
    reparam.has_post_rescaling = False
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x"]
    reparam.offsets = {"x": 1.0}
    reparam.boundary_inversion = {"x": "split"}
    x_prime = numpy_array_to_live_points(
        np.array([(-1.0,), (2.0,)]), reparam.prime_parameters
    )
    inversion_out = numpy_array_to_live_points(
        np.array([(1.0,), (2.0,)]), reparam.parameters
    )
    x_in = np.zeros(
        [
            2,
        ],
        dtype=get_dtype(reparam.parameters),
    )
    log_j = np.zeros(x_prime.size)

    x_ex = inversion_out
    x_prime_ex = x_prime
    log_j_ex = np.array([0, 0.5])

    reparam._reverse_inversion = MagicMock(
        return_value=(x_ex, x_prime_ex, log_j_ex)
    )

    x_out, x_prime_out, log_j_out = RescaleToBounds.inverse_reparameterise(
        reparam,
        x_in,
        x_prime,
        log_j,
    )

    assert_structured_arrays_equal(
        reparam._reverse_inversion.call_args_list[0][0][0], x_in
    )
    assert_structured_arrays_equal(
        reparam._reverse_inversion.call_args_list[0][0][1], x_prime
    )
    np.testing.assert_array_equal(
        reparam._reverse_inversion.call_args_list[0][0][2], log_j
    )

    assert_structured_arrays_equal(x_out, x_ex)
    assert_structured_arrays_equal(x_prime_out, x_prime_ex)
    np.testing.assert_array_equal(log_j_out, log_j_ex)


def test_reparameterise_pre_post_rescaling(reparam):
    """Test the reparameterise function with pre and post rescaling"""
    reparam.has_pre_rescaling = True
    reparam.has_post_rescaling = True
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.offsets = {"x": 0.0}
    reparam.boundary_inversion = {}
    x = numpy_array_to_live_points(
        np.array([(1.0,), (2.0,)]), reparam.parameters
    )
    x_prime_in = np.zeros(
        [
            2,
        ],
        dtype=get_dtype(reparam.prime_parameters),
    )
    x_prime_val = np.array([4.0, 8.0])
    log_j = np.zeros(x.size)

    reparam._rescale_to_bounds = MagicMock(
        return_value=(x_prime_val, np.array([0, 0.5]))
    )
    reparam.pre_rescaling = MagicMock(
        return_value=(np.array([0.5, 1.0]), np.array([0.5, 0.5]))
    )
    reparam.post_rescaling = MagicMock(
        return_value=(np.array([1.0, 1.5]), np.array([2.0, 3.0]))
    )

    x_out, x_prime_out, log_j_out = RescaleToBounds.reparameterise(
        reparam, x, x_prime_in, log_j
    )

    np.testing.assert_array_equal(
        np.array([0.5, 1.0]),
        reparam._rescale_to_bounds.call_args_list[0][0][0],
    )

    assert reparam._rescale_to_bounds.call_args_list[0][0][1] == "x"

    np.testing.assert_array_equal(
        reparam.pre_rescaling.call_args_list[0][0][0],
        np.array([1.0, 2.0]),
    )
    # x_prime gets replaces so change check inputs to the function
    reparam.post_rescaling.assert_called_once()

    assert_structured_arrays_equal(x, x_out)
    np.testing.assert_array_equal(x_prime_out["x_prime"], np.array([1.0, 1.5]))
    np.testing.assert_array_equal(log_j_out, np.array([2.5, 4.0]))


def test_inverse_reparameterise_pre_post_rescaling(reparam):
    """Test the inverse_reparameterise function with pre and post rescaling"""
    reparam.has_pre_rescaling = True
    reparam.has_post_rescaling = True
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.offsets = {"x": 0.0}
    reparam.boundary_inversion = {}
    x_prime = numpy_array_to_live_points(
        np.array([(1.0,), (2.0,)]), reparam.prime_parameters
    )
    x_in = np.zeros(
        [
            2,
        ],
        dtype=get_dtype(reparam.parameters),
    )
    x_val = np.array([4.0, 8.0])
    log_j = np.zeros(x_prime.size)

    reparam._inverse_rescale_to_bounds = MagicMock(
        return_value=(x_val, np.array([0, 0.5]))
    )
    reparam.pre_rescaling_inv = MagicMock(
        return_value=(np.array([0.5, 1.0]), np.array([0.5, 0.5]))
    )
    reparam.post_rescaling_inv = MagicMock(
        return_value=(np.array([1.0, 1.5]), np.array([2.0, 3.0]))
    )

    x_out, x_prime_out, log_j_out = RescaleToBounds.inverse_reparameterise(
        reparam, x_in, x_prime, log_j
    )

    np.testing.assert_array_equal(
        np.array([0.5, 1.0]),
        reparam._inverse_rescale_to_bounds.call_args_list[0][0][0],
    )

    assert reparam._inverse_rescale_to_bounds.call_args_list[0][0][1] == "x"

    np.testing.assert_array_equal(
        reparam.post_rescaling_inv.call_args_list[0][0][0],
        x_prime["x_prime"],
    )
    reparam.pre_rescaling_inv.assert_called_once()
    # x_prime gets replaces so change check inputs to the function
    reparam.post_rescaling_inv.assert_called_once()

    assert_structured_arrays_equal(x_prime, x_prime_out)
    np.testing.assert_array_equal(x_out["x"], np.array([0.5, 1.0]))
    np.testing.assert_array_equal(log_j_out, np.array([2.5, 4.0]))


def test_apply_inversion_detect_edge(reparam):
    """Assert detect edge is called with the correct arguments"""
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.offsets = {"x": 1.0}
    reparam._edges = {"x": None}
    reparam.detect_edges_kwargs = {"allowed_bounds": ["lower"]}
    reparam.bounds = {"x": [0, 5]}
    reparam.rescale_bounds = {"x": [0, 1]}

    x = numpy_array_to_live_points(np.array([1, 2]), ["x"])
    x_prime = numpy_array_to_live_points(np.array([3, 4]), ["x_prime"])
    log_j = np.zeros(2)

    with patch(
        "nessai.reparameterisations.rescale.detect_edge", return_value=False
    ) as mock_fn:

        _ = RescaleToBounds._apply_inversion(
            reparam, x, x_prime, log_j, "x", "x_prime", False, test=True
        )

    reparam.update_prime_prior_bounds.assert_called_once()
    mock_fn.assert_called_once_with(
        x_prime["x_prime"],
        test=True,
        allowed_bounds=["lower"],
    )

    assert reparam._edges == {"x": False}


def test_apply_inversion_not_applied(reparam):
    """Assert the apply inversion works correctly"""
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.offsets = {"x": 1.0}
    reparam._edges = {"x": False}
    reparam.bounds = {"x": [0, 5]}
    reparam.rescale_bounds = {"x": [0, 1]}

    x_val = np.array([[1], [2]])
    x_prime = numpy_array_to_live_points(x_val, ["x_prime"])
    x = numpy_array_to_live_points(np.array([3, 4]), ["x"])
    log_j = np.zeros(2)

    with patch(
        "nessai.reparameterisations.rescale.rescale_minus_one_to_one",
        side_effect=lambda x, *args, **kwargs: (x, np.array([5, 6])),
    ) as f:
        x_out, x_prime_out, log_j_out = RescaleToBounds._apply_inversion(
            reparam,
            x,
            x_prime,
            log_j,
            "x",
            "x_prime",
            False,
        )

    assert f.call_args_list[0][1] == {"xmin": 0, "xmax": 5}
    # Should be output of rescaling minus offset
    np.testing.assert_array_equal(x_prime_out["x_prime"], np.array([0, 1]))
    # x_prime should be the same
    assert x_out is x
    # Jacobian should just include jacobian from rescaling
    np.testing.assert_array_equal(log_j_out, np.array([5, 6]))


def test_apply_inversion_split(reparam):
    """Assert apply inversion with split works as intended.

    Also tests "upper" setting.
    """
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.offsets = {"x": 1.0}
    reparam._edges = {"x": "upper"}
    reparam.bounds = {"x": [0, 5]}
    reparam.rescale_bounds = {"x": [0, 1]}
    reparam.boundary_inversion = {"x": "split"}

    x_val = np.array([[1.2], [1.7]])
    x_prime = numpy_array_to_live_points(x_val, ["x_prime"])
    x = numpy_array_to_live_points(np.array([3, 4]), ["x"])
    log_j = np.zeros(2)

    with patch(
        "numpy.random.choice", return_value=np.array([1])
    ) as rnd, patch(
        "nessai.reparameterisations.rescale.rescale_zero_to_one",
        side_effect=lambda x, *args: (x, np.array([5, 6])),
    ) as f:
        x_out, x_prime_out, log_j_out = RescaleToBounds._apply_inversion(
            reparam,
            x,
            x_prime,
            log_j,
            "x",
            "x_prime",
            False,
        )

    rnd.assert_called_once_with(2, 1, replace=False)
    assert f.call_args_list[0][0][1] == 0.0
    assert f.call_args_list[0][0][2] == 5.0
    # Output should be x_val minus offset
    # Then 1 - that for 'upper'
    # Then *= -1 with the output of rnd
    np.testing.assert_array_almost_equal(
        x_prime_out["x_prime"],
        np.array([0.8, -0.3]),
        decimal=10,
    )
    # x should be the same
    assert x_out is x
    # Jacobian should just include jacobian from rescaling
    np.testing.assert_array_equal(log_j_out, np.array([5, 6]))


@pytest.mark.parametrize(
    "inv_type, compute_radius",
    [("duplicate", False), ("duplicate", True), ("split", True)],
)
def test_apply_inversion_duplicate(reparam, inv_type, compute_radius):
    """Assert apply inversion with duplicate works as intended.

    This test also covers compute_radius=True
    """
    reparam.parameters = ["x", "y"]
    reparam.prime_parameters = ["x_prime", "y"]
    reparam.offsets = {"x": 1.0}
    reparam._edges = {"x": "lower"}
    reparam.bounds = {"x": [0, 5]}
    reparam.rescale_bounds = {"x": [0, 1]}
    reparam.boundary_inversion = {"x": inv_type}

    x_val = np.array([[1.2, 1.0], [1.7, 2.0]])
    x_prime = numpy_array_to_live_points(x_val, ["x_prime", "y"])
    x = numpy_array_to_live_points(np.array([[1, 2], [3, 4]]), ["x", "y"])
    log_j = np.zeros(2)

    with patch("numpy.random.choice") as rnd, patch(
        "nessai.reparameterisations.rescale.rescale_zero_to_one",
        side_effect=lambda x, *args: (x, np.array([5, 6])),
    ) as f:
        x_out, x_prime_out, log_j_out = RescaleToBounds._apply_inversion(
            reparam,
            x,
            x_prime,
            log_j,
            "x",
            "x_prime",
            compute_radius,
        )

    rnd.assert_not_called()
    assert f.call_args_list[0][0][1] == 0.0
    assert f.call_args_list[0][0][2] == 5.0
    # Output should be x_val minus offset
    # Then duplicated
    np.testing.assert_array_almost_equal(
        x_prime_out["x_prime"],
        np.array([0.2, 0.7, -0.2, -0.7]),
        decimal=10,
    )
    np.testing.assert_array_almost_equal(
        x_prime_out["y"],
        np.array([1.0, 2.0, 1.0, 2.0]),
        decimal=10,
    )
    # x should be the same but duplicated
    assert_structured_arrays_equal(x_out, np.concatenate([x, x]))
    # Jacobian should just include jacobian from rescaling but duplicated
    np.testing.assert_array_equal(log_j_out, np.array([5, 6, 5, 6]))


def test_reverse_inversion(reparam):
    """Assert the reverse inversion works correctly"""
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.offsets = {"x": 1.0}
    reparam._edges = {"x": "upper"}
    reparam.bounds = {"x": [0, 5]}

    x_val = np.array([[-0.7], [0.4]])
    x = numpy_array_to_live_points(x_val, ["x"])
    x_prime = numpy_array_to_live_points(np.array([3, 4]), ["x_prime"])
    log_j = np.zeros(2)

    # Return the same value to check that the negative values are handled
    # correctly
    with patch(
        "nessai.reparameterisations.rescale.inverse_rescale_zero_to_one",
        side_effect=lambda x, *args: (x, np.array([5, 6])),
    ) as f:
        x_out, x_prime_out, log_j_out = RescaleToBounds._reverse_inversion(
            reparam,
            x,
            x_prime,
            log_j,
            "x",
            "x_prime",
        )

    assert f.call_args_list[0][0][1] == 0.0
    assert f.call_args_list[0][0][2] == 5.0
    # Should be output of rescaling minus offset
    np.testing.assert_array_equal(x_out["x"], np.array([1.3, 1.6]))
    # x_prime should be the same
    assert x_prime_out is x_prime
    # Jacobian should just include jacobian from rescaling
    np.testing.assert_array_equal(log_j_out, np.array([5, 6]))


def test_reverse_inversion_not_applied(reparam):
    """Assert the reverse inversion works correctly"""
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.offsets = {"x": 1.0}
    reparam._edges = {"x": False}
    reparam.bounds = {"x": [0, 5]}

    x_val = np.array([[1], [2]])
    x = numpy_array_to_live_points(x_val, ["x"])
    x_prime = numpy_array_to_live_points(np.array([3, 4]), ["x_prime"])
    log_j = np.zeros(2)

    with patch(
        "nessai.reparameterisations.rescale.inverse_rescale_minus_one_to_one",
        side_effect=lambda x, *args, **kwargs: (x, np.array([5, 6])),
    ) as f:
        x_out, x_prime_out, log_j_out = RescaleToBounds._reverse_inversion(
            reparam,
            x,
            x_prime,
            log_j,
            "x",
            "x_prime",
        )

    assert f.call_args_list[0][1] == {"xmin": 0, "xmax": 5}
    # Should be output of rescaling minus offset
    np.testing.assert_array_equal(x_out["x"], np.array([2, 3]))
    # x_prime should be the same
    assert x_prime_out is x_prime
    # Jacobian should just include jacobian from rescaling
    np.testing.assert_array_equal(log_j_out, np.array([5, 6]))


@pytest.mark.parametrize(
    "rescale_bounds", [None, [0, 1], {"x": [0, 1], "y": [-1, 1]}]
)
@pytest.mark.integration_test
def test_rescale_bounds(reparameterisation, is_invertible, rescale_bounds):
    """Test the different options for rescale to bounds"""
    reparam = reparameterisation({"rescale_bounds": rescale_bounds})
    if rescale_bounds is None:
        rescale_bounds = {p: [-1, 1] for p in reparam.parameters}
    elif isinstance(rescale_bounds, list):
        rescale_bounds = {p: rescale_bounds for p in reparam.parameters}

    assert reparam.rescale_bounds == rescale_bounds
    assert is_invertible(reparam)


@pytest.mark.parametrize(
    "boundary_inversion",
    [False, True, ["x"], {"x": "split"}, {"x": "duplicate"}],
)
@pytest.mark.integration_test
def test_boundary_inversion(
    reparameterisation, is_invertible, boundary_inversion
):
    """Test the different options for rescale to bounds"""
    reparam = reparameterisation({"boundary_inversion": boundary_inversion})

    assert is_invertible(reparam)


@pytest.mark.integration_test
def test_update_prime_prior_bounds_integration():
    """Assert the prime prior bounds are correctly computed"""
    rescaling = (
        lambda x: (x / 2, np.zeros_like(x)),
        lambda x: (2 * x, np.zeros_like(x)),
    )
    reparam = RescaleToBounds(
        parameters=["x"],
        prior_bounds=[1000, 1001],
        prior="uniform",
        pre_rescaling=rescaling,
        offset=True,
    )
    np.testing.assert_equal(reparam.offsets["x"], 500.25)
    np.testing.assert_array_equal(reparam.prior_bounds["x"], [1000, 1001])
    np.testing.assert_array_equal(reparam.pre_prior_bounds["x"], [500, 500.5])
    np.testing.assert_array_equal(reparam.bounds["x"], [-0.25, 0.25])
    np.testing.assert_array_equal(
        reparam.prime_prior_bounds["x_prime"], [-1, 1]
    )

    x_prime = numpy_array_to_live_points(
        np.array([[-2], [-1], [0.5], [1], [10]]), ["x_prime"]
    )
    log_prior = reparam.x_prime_log_prior(x_prime)
    expected = np.array([-np.inf, 0, 0, 0, -np.inf])
    np.testing.assert_equal(log_prior, expected)


@pytest.mark.integration_test
def test_pre_rescaling_integration(is_invertible, model):
    """Test the pre-scaling feature"""

    def forward(x):
        return np.log(x), -np.log(x)

    def inv(x):
        return np.exp(x), x.copy()

    reparam = RescaleToBounds(
        parameters="x",
        prior_bounds={"x": [1.0, np.e]},
        pre_rescaling=(forward, inv),
        rescale_bounds=[-1.0, 1.0],
    )

    x = numpy_array_to_live_points(
        np.array([[1.0], [np.e**0.5], [2.0], [np.e]]), ["x"]
    )
    x_prime = empty_structured_array(x.size, ["x_prime"])
    log_j = np.zeros(x.size)

    x_out, x_prime_out, log_j_out = reparam.reparameterise(x, x_prime, log_j)

    assert_structured_arrays_equal(x_out, x)
    np.testing.assert_allclose(
        x_prime_out["x_prime"],
        np.array([-1, 0.0, 2 * np.log(2) - 1, 1]),
        rtol=rtol,
        atol=atol,
    )
    np.testing.assert_allclose(
        log_j_out,
        -np.log(x["x"]) + np.log(2),
        rtol=rtol,
        atol=atol,
    )

    x_in = empty_structured_array(x_prime_out.size, ["x"])
    log_j = np.zeros(x.size)
    x_out, x_prime_final, log_j_final = reparam.inverse_reparameterise(
        x_in, x_prime_out, log_j
    )

    np.testing.assert_allclose(
        log_j_final,
        np.log(x_out["x"]) - np.log(2),
        rtol=rtol,
        atol=atol,
    )

    np.testing.assert_allclose(x_out["x"], x["x"], rtol=rtol, atol=atol)
    assert_structured_arrays_equal(x_prime_final, x_prime_out)
    np.testing.assert_allclose(log_j_final, -log_j_out, rtol=rtol, atol=atol)

    # Trick to get a 1d model to test only x
    model._names.remove("y")
    model._bounds = {"x": [1.0, np.e]}
    assert is_invertible(reparam, model=model, decimal=12)


@pytest.mark.integration_test
def test_update_integration(model):
    """Assert edges and bounds are updated"""

    x = model.new_point(2)

    new_bounds = {"x": [np.min(x["x"]), np.max(x["x"])]}

    reparam = RescaleToBounds(
        parameters="x",
        update_bounds=True,
        boundary_inversion=True,
        detect_edges=True,
        prior_bounds=model.bounds["x"],
    )

    reparam._edges = {"x": "lower"}
    reparam.bounds = {"x": [-100, 100]}

    reparam.update(x)

    assert reparam._edges == {"x": None}
    assert reparam.bounds == new_bounds


@pytest.mark.integration_test
def test_update_integration_no_update(model):
    """Assert the bounds and edges are not updated if disabled."""
    x = model.new_point(2)

    reparam = RescaleToBounds(
        parameters="x",
        update_bounds=False,
        boundary_inversion=False,
        detect_edges=False,
        prior_bounds=model.bounds["x"],
    )

    reparam.update(x)

    assert reparam._edges is None
    np.testing.assert_array_equal(reparam.bounds["x"], model.bounds["x"])


@pytest.mark.parametrize(
    "kwargs, decimal",
    [
        (dict(post_rescaling="logit", update_bounds=False), 10),
        (
            dict(
                post_rescaling="log",
                update_bounds=False,
            ),
            14,
        ),
        (dict(update_bounds=False), None),
        (dict(update_bounds=False, boundary_inversion=True), None),
        (dict(boundary_inversion=["x"]), None),
    ],
)
@pytest.mark.integration_test
def test_is_invertible_general_config(is_invertible, model, kwargs, decimal):
    """Test the invertibility of the reparameterisation

    General tests that don't check specific attributes but can check
    combinations.
    """
    if decimal is None:
        decimal = 16
    default_kwargs = dict(
        parameters=model.names,
        prior_bounds=model.bounds,
    )
    default_kwargs.update(kwargs)
    reparam = RescaleToBounds(**default_kwargs)
    assert is_invertible(reparam, decimal=decimal)



# -*- coding: utf-8 -*-
"""
Implementation of MaskedAutoregressiveFlow.
"""
import logging

from torch.nn import functional as F

from glasflow.nflows.distributions.normal import StandardNormal
from glasflow.nflows.transforms.autoregressive import (
    MaskedAffineAutoregressiveTransform,
)
from glasflow.nflows.transforms.base import CompositeTransform
from glasflow.nflows.transforms.normalization import BatchNorm
from glasflow.nflows.transforms.permutations import (
    RandomPermutation,
    ReversePermutation,
)

from .base import NFlow

logger = logging.getLogger(__name__)


class MaskedAutoregressiveFlow(NFlow):
    """Autoregressive flow with masked coupling transforms.

    Based on the implementation from nflows: \
        https://github.com/bayesiains/nflows/blob/master/nflows/flows/autoregressive.py
        but also included context features.

    Parameters
    ----------
    features : int
        Number of features (dimensions) in the data space
    hidden_features : int
        Number of neurons per layer in each neural network
    num_layers : int
        Number of coupling transformations
    num_blocks_per_layer : int
        Number of layers (or blocks for resnet) per neural network for
        each coupling transform
    context_features : int, optional
        Number of context (conditional) parameters.
    use_residual_blocks : bool, optional
        Use residual blocks in the MADE network.
    use_random_masks : bool, optional
        Use random masks in the MADE network.
    use_random_permutation : bool, optional
        Use a random permutation instead of the default reverse permutation.
    activation : function, optional
        Activation function implemented in torch.
    dropout_probability : float, optional
        Dropout probability used in each layer of the neural network
    batch_norm_within_layers : bool, optional
       Enable or disable batch norm within the neural network for each coupling
       transform
    batch_norm_between_layers : bool, optional
       Enable or disable batch norm between coupling transforms
    """

    def __init__(
        self,
        features,
        hidden_features,
        num_layers,
        num_blocks_per_layer,
        context_features=None,
        use_residual_blocks=True,
        use_random_masks=False,
        use_random_permutations=False,
        activation=F.relu,
        dropout_probability=0.0,
        batch_norm_within_layers=False,
        batch_norm_between_layers=False,
    ):

        if use_random_permutations:
            permutation_constructor = RandomPermutation
        else:
            permutation_constructor = ReversePermutation

        layers = []
        for _ in range(num_layers):
            layers.append(permutation_constructor(features))
            layers.append(
                MaskedAffineAutoregressiveTransform(
                    features=features,
                    hidden_features=hidden_features,
                    context_features=context_features,
                    num_blocks=num_blocks_per_layer,
                    use_residual_blocks=use_residual_blocks,
                    random_mask=use_random_masks,
                    activation=activation,
                    dropout_probability=dropout_probability,
                    use_batch_norm=batch_norm_within_layers,
                )
            )
            if batch_norm_between_layers:
                layers.append(BatchNorm(features))

        super().__init__(
            transform=CompositeTransform(layers),
            distribution=StandardNormal([features]),
        )



# -*- coding: utf-8 -*-
"""
Specific proposal methods for sampling gravitational-wave models.
"""
import logging

from ..proposal import FlowProposal
from ..proposal.augmented import AugmentedFlowProposal

from .reparameterisations import get_gw_reparameterisation

logger = logging.getLogger(__name__)


class GWFlowProposal(FlowProposal):
    """Wrapper for FlowProposal that has defaults for CBC-PE"""

    aliases = {
        "chirp_mass": ("mass", None),
        "mass_ratio": ("mass_ratio", None),
        "ra": ("sky-ra-dec", ["DEC", "dec", "Dec"]),
        "dec": ("sky-ra-dec", ["RA", "ra"]),
        "azimuth": ("sky-az-zen", ["zenith", "zen", "Zen", "Zenith"]),
        "zenith": ("sky-az-zen", ["azimuth", "az", "Az", "Azimuth"]),
        "theta_1": ("angle-sine", None),
        "theta_2": ("angle-sine", None),
        "tilt_1": ("angle-sine", None),
        "tilt_2": ("angle-sine", None),
        "theta_jn": ("angle-sine", None),
        "iota": ("angle-sine", None),
        "phi_jl": ("angle-2pi", None),
        "phi_12": ("angle-2pi", None),
        "phase": ("angle-2pi", None),
        "psi": ("angle-pi", None),
        "geocent_time": ("time", None),
        "time_jitter": ("periodic", None),
        "a_1": ("default", None),
        "a_2": ("default", None),
        "chi_1": ("default", None),
        "chi_2": ("default", None),
        "luminosity_distance": ("distance", None),
    }
    """
    Dictionary of aliases used to determine the default reparameterisations
    for common gravitational-wave parameters.
    """
    use_default_reparameterisations = True
    """
    GW specific reparameterisations will be used by default. This is different
    to the parent class where they are disabled by default.
    """

    def get_reparameterisation(self, reparameterisation):
        """Function to get reparameterisations that checks GW defaults and
        aliases
        """
        return get_gw_reparameterisation(reparameterisation)

    def add_default_reparameterisations(self):
        """
        Add default reparameterisations for parameters that have not been
        specified.
        """
        parameters = [
            n
            for n in self.model.names
            if n not in self._reparameterisation.parameters
        ]
        logger.info(f"Adding default reparameterisations for {parameters}")

        for p in parameters:
            logger.debug(f"Trying to add reparameterisation for {p}")
            if p in self._reparameterisation.parameters:
                logger.debug(f"Parameter {p} is already included")
                continue
            name, extra_params = self.aliases.get(p.lower(), (None, None))
            if name is None:
                logger.debug(f"{p} is not a known GW parameter")
                continue
            if extra_params is not None:
                p = [p] + [ep for ep in extra_params if ep in self.model.names]
            else:
                p = [p]
            prior_bounds = {k: self.model.bounds[k] for k in p}
            reparam, kwargs = get_gw_reparameterisation(name)
            logger.info(
                f"Adding reparameterisation {reparam.__name__} for {p} "
                f"with config: {kwargs}"
            )
            self._reparameterisation.add_reparameterisation(
                reparam(parameters=p, prior_bounds=prior_bounds, **kwargs)
            )


class AugmentedGWFlowProposal(AugmentedFlowProposal, GWFlowProposal):
    """Augmented version of GWFlowProposal.

    See :obj:`~nessai.proposal.augmented.AugmentedFlowProposal` and
    :obj:`~nessai.gw.proposal.GWFlowPropsosal`
    """

    pass



# -*- coding: utf-8 -*-
"""Test methods related to initialising and resuming the proposal method"""
import numpy as np
import os
import pytest
from unittest.mock import patch, MagicMock

from nessai.proposal import FlowProposal


def test_init(model):
    """Test init with some kwargs"""
    fp = FlowProposal(model, poolsize=1000)
    assert fp.model == model


@pytest.mark.parametrize(
    "value, expected", [(True, True), (False, False), (None, False)]
)
def test_init_use_default_reparams(model, proposal, value, expected):
    """Assert use_default_reparameterisations is set correctly"""
    proposal.use_default_reparameterisations = False
    FlowProposal.__init__(
        proposal, model, poolsize=10, use_default_reparameterisations=value
    )
    assert proposal.use_default_reparameterisations is expected


@pytest.mark.parametrize("ef, fuzz", [(2.0, 3.0**0.5), (False, 2.0)])
def test_initialise(tmpdir, proposal, ef, fuzz):
    """Test the initialise method"""
    p = tmpdir.mkdir("test")
    proposal.output = os.path.join(p, "output")
    proposal.rescaled_dims = 2
    proposal.expansion_fraction = ef
    proposal.fuzz = 2.0
    proposal.flow_config = {"model_config": {}}
    proposal.set_rescaling = MagicMock()
    proposal.verify_rescaling = MagicMock()
    proposal.update_flow_config = MagicMock()
    proposal.configure_constant_volume = MagicMock()
    fm = MagicMock()
    fm.initialise = MagicMock()

    with patch("nessai.proposal.flowproposal.FlowModel", new=fm) as mock_fm:
        FlowProposal.initialise(proposal)

    proposal.set_rescaling.assert_called_once()
    proposal.verify_rescaling.assert_called_once()
    proposal.update_flow_config.assert_called_once()
    proposal.configure_constant_volume.assert_called_once()
    mock_fm.assert_called_once_with(
        config=proposal.flow_config, output=proposal.output
    )
    proposal.flow.initialise.assert_called_once()
    assert proposal.populated is False
    assert proposal.initialised
    assert proposal.fuzz == fuzz
    assert os.path.exists(os.path.join(p, "output"))


def test_resume(proposal):
    """Test the resume method."""
    from numpy import array, array_equal

    proposal.initialise = MagicMock()
    proposal.mask = [1, 0]
    proposal.update_bounds = False
    proposal.weights_file = None
    model = MagicMock()
    with patch("nessai.proposal.base.Proposal.resume") as mock:
        FlowProposal.resume(proposal, model, {"model_config": {"kwargs": {}}})
    mock.assert_called_once_with(model)
    proposal.initialise.assert_called_once()
    assert array_equal(
        proposal.flow_config["model_config"]["kwargs"]["mask"], array([1, 0])
    )


@patch("os.path.exists", return_value=True)
def test_resume_w_weights(osexist, proposal):
    """Test the resume method with weights"""
    proposal.initialise = MagicMock()
    proposal.flow = MagicMock()
    proposal.mask = None
    proposal.update_bounds = False
    proposal.weights_file = None
    model = MagicMock()
    with patch("nessai.proposal.base.Proposal.resume") as mock:
        FlowProposal.resume(proposal, model, {}, weights_file="weights.pt")
    mock.assert_called_once_with(model)
    osexist.assert_called_once_with("weights.pt")
    proposal.initialise.assert_called_once()
    proposal.flow.reload_weights.assert_called_once_with("weights.pt")


@pytest.mark.parametrize("data", [[1], None])
@pytest.mark.parametrize("count", [0, 1])
def test_resume_w_update_bounds(proposal, data, count):
    """Test the resume method with update bounds"""
    proposal.initialise = MagicMock()
    proposal.flow = MagicMock()
    proposal.mask = None
    proposal.update_bounds = True
    proposal.weights_file = None
    proposal.training_data = data
    proposal.training_count = count
    proposal.check_state = MagicMock()
    model = MagicMock()
    if count and data is None:
        with pytest.raises(RuntimeError) as excinfo, patch(
            "nessai.proposal.base.Proposal.resume"
        ) as mock:
            FlowProposal.resume(proposal, model, {})
        assert "Could not resume" in str(excinfo.value)
    else:
        with patch("nessai.proposal.base.Proposal.resume") as mock:
            FlowProposal.resume(proposal, model, {})
        if data:
            proposal.check_state.assert_called_once_with(data)
    mock.assert_called_once_with(model)


@pytest.mark.parametrize("populated", [False, True])
@pytest.mark.parametrize("mask", [None, [1, 0]])
def test_get_state(proposal, populated, mask):
    """Test the get state method used for pickling the proposal.

    Tests cases where the proposal is and isn't populated.
    """

    proposal.populated = populated
    proposal.indices = [1, 2]
    proposal._reparameterisation = MagicMock()
    proposal.model = MagicMock()
    proposal._flow_config = {}
    proposal.initialised = True
    proposal.flow = MagicMock()
    proposal.flow.weights_file = "file"
    proposal._draw_func = lambda x: x

    if mask is not None:
        proposal.flow.model_config = {"kwargs": {"mask": mask}}

    state = FlowProposal.__getstate__(proposal)

    assert state["resume_populated"] is populated
    assert state["initialised"] is False
    assert state["weights_file"] == "file"
    assert state["mask"] is mask
    assert "_reparameterisation" not in state
    assert "model" not in state
    assert "flow" not in state
    assert "_flow_config" not in state


@pytest.mark.integration_test
@pytest.mark.parametrize("reparameterisation", [False, True])
@pytest.mark.parametrize("init", [False, True])
@pytest.mark.parametrize(
    "latent_prior", ["truncated_gaussian", "uniform_nball", "gaussian"]
)
def test_resume_pickle(model, tmpdir, reparameterisation, init, latent_prior):
    """Test pickling and resuming the proposal.

    Tests both with and without reparameterisations and before and after
    initialise has been called.
    """
    import pickle

    output = tmpdir.mkdir("test_integration")
    if reparameterisation:
        reparameterisations = {"default": {"parameters": model.names}}
    else:
        reparameterisations = None

    if latent_prior != "truncated_gaussian":
        constant_volume_mode = False
    else:
        constant_volume_mode = True

    proposal = FlowProposal(
        model,
        poolsize=1000,
        plot=False,
        expansion_fraction=1,
        output=output,
        reparameterisations=reparameterisations,
        latent_prior=latent_prior,
        constant_volume_mode=constant_volume_mode,
    )
    if init:
        proposal.initialise()

    proposal.mask = None
    proposal.resume_populated = False

    proposal_data = pickle.dumps(proposal)
    proposal_re = pickle.loads(proposal_data)
    proposal_re.resume(model, {})

    assert proposal._plot_pool == proposal_re._plot_pool
    assert proposal._plot_training == proposal_re._plot_training

    if init:
        assert proposal.fuzz == proposal_re.fuzz
        assert proposal.rescaled_names == proposal_re.rescaled_names


def test_reset(proposal):
    """Test reset method"""
    proposal.x = 1
    proposal.samples = 2
    proposal.populated = True
    proposal.populated_count = 10
    FlowProposal.reset(proposal)
    assert proposal.x is None
    assert proposal.samples is None
    assert proposal.populated is False
    assert proposal.populated_count == 0
    assert proposal.r is np.nan
    assert proposal.alt_dist is None
    assert proposal._checked_population


@pytest.mark.timeout(60)
@pytest.mark.integration_test
@pytest.mark.parametrize(
    "latent_prior", ["truncated_gaussian", "uniform_nball", "gaussian"]
)
def test_reset_integration(tmpdir, model, latent_prior):
    """Test reset method iteration with other methods"""
    config = dict(
        model_config=dict(
            n_neurons=1,
            n_blocks=1,
            n_layers=1,
            kwargs=dict(
                batch_norm_between_layers=False,
                linear_transform=None,
            ),
        )
    )
    output = str(tmpdir.mkdir("reset_integration"))
    poolsize = 2
    drawsize = 100
    if latent_prior != "truncated_gaussian":
        constant_volume_mode = False
    else:
        constant_volume_mode = True
    proposal = FlowProposal(
        model,
        output=output,
        flow_config=config,
        plot=False,
        poolsize=poolsize,
        drawsize=drawsize,
        expansion_fraction=None,
        latent_prior=latent_prior,
        constant_volume_mode=constant_volume_mode,
    )

    modified_proposal = FlowProposal(
        model,
        output=output,
        flow_config=config,
        plot=False,
        poolsize=poolsize,
        drawsize=drawsize,
        expansion_fraction=None,
        latent_prior=latent_prior,
        constant_volume_mode=constant_volume_mode,
    )
    proposal.initialise()
    modified_proposal.initialise()

    modified_proposal.populate(model.new_point())
    modified_proposal.reset()

    # attributes that should be different
    ignore = [
        "population_time",
    ]

    d1 = proposal.__getstate__()
    d2 = modified_proposal.__getstate__()

    for key in ignore:
        d1.pop(key)
        d2.pop(key)

    assert d1 == d2



# -*- coding: utf-8 -*-
"""
Functions related to creating live points and converting to other common
data-types.
"""
import logging

import numpy as np
from numpy.lib import recfunctions as rfn

from . import config

logger = logging.getLogger(__name__)


def add_extra_parameters_to_live_points(
    parameters,
    default_values=None,
):
    """Add extra parameters to the live points dtype.

    Extra parameters will be included in the live points dtype that is used
    for constructing/converting to/from live points.

    Parameters
    ----------
    parameters: list
        List of parameters to add.
    default_values: Optional[Union[List, Tuple]]
        List of default values for each parameters. If not specified, default
        values will be set to based on :code: `DEFAULT_FLOAT_VALUE` in
        :code:`nessai.config`.
    """
    if default_values is None:
        default_values = len(parameters) * (
            config.livepoints.default_float_value,
        )
    else:
        default_values = tuple(default_values)
    for p, dv in zip(parameters, default_values):
        if p not in config.livepoints.extra_parameters:
            config.livepoints.extra_parameters.append(p)
            config.livepoints.extra_parameters_defaults = (
                config.livepoints.extra_parameters_defaults + (dv,)
            )
            config.livepoints.extra_parameters_dtype.append(
                config.livepoints.default_float_dtype
            )
        else:
            logger.warning(
                f"Extra parameter `{p}` has already been added. Skipping."
                "Call `reset_extra_live_points_parameters` to reset the values"
                " and add this parameter."
            )

    logger.debug(
        "Updated non-sampling parameters: "
        f"{config.livepoints.non_sampling_parameters}"
    )
    logger.debug(
        "Updated defaults for non-sampling parameters: "
        f"{config.livepoints.non_sampling_defaults}"
    )
    config.livepoints.reset_properties()


def reset_extra_live_points_parameters():
    """Reset the extra live points parameters."""
    logger.debug("Resetting extra parameters")
    config.livepoints.reset()


def get_dtype(names, array_dtype=None, non_sampling_parameters=True):
    """
    Get a list of tuples containing the dtypes for the structured array

    Parameters
    ----------
    names : list of str
        Names of parameters
    array_dtype : Optional[str]
        dtype to use
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    numpy.dtype
        A instance of :code:`numpy.dtype`.
    """
    if array_dtype is None:
        array_dtype = config.livepoints.default_float_dtype
    dtype = [(n, array_dtype) for n in names]
    if non_sampling_parameters:
        dtype += list(
            zip(
                config.livepoints.non_sampling_parameters,
                config.livepoints.non_sampling_dtype,
            )
        )
    return np.dtype(dtype)


def empty_structured_array(
    n, names=None, dtype=None, non_sampling_parameters=True
):
    """Get an empty structured array with the extra parameters initialised.

    Parameters
    ----------
    n : int
        Length of the structured array
    dtype : Optional[list]
        Dtype to use. Must contain the non-sampling parameters.
    names : Optional[list]
        Names of fields (excluding non-sampling parameters) to construct the
        dtype. Must be specified if :code:`dtype` is not specified.
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    np.ndarray
        Structured array with the all parameters initialised to their
        default values.
    """
    if dtype is None:
        dtype = get_dtype(
            names, non_sampling_parameters=non_sampling_parameters
        )
    else:
        dtype = np.dtype(dtype)
        names = [
            nm
            for nm in dtype.names
            if nm not in config.livepoints.non_sampling_parameters
        ]
    struct_array = np.empty((n), dtype=dtype)
    if n == 0:
        return struct_array
    struct_array[names] = config.livepoints.default_float_value
    if non_sampling_parameters:
        try:
            for nm, v in zip(
                config.livepoints.non_sampling_parameters,
                config.livepoints.non_sampling_defaults,
            ):
                struct_array[nm] = v
        except ValueError:
            raise ValueError(
                "Could not create empty structured array. Maybe the "
                "non-sampling parameters are missing?"
            )
    return struct_array


def live_points_to_array(live_points, names=None, copy=False):
    """
    Converts live points to unstructured arrays for training.

    Parameters
    ----------
    live_points : structured_array
        Structured array of live points
    names : list of str or None
        If None all fields in the structured array are added to the dictionary
        else only those included in the list are added.
    copy : bool
        If true, returns a copy. If false, returns a view. See numpy
        documentation for
        :code:`numpy.lib.recfunctions.structured_to_unstructured` for more
        details.

    Returns
    -------
    np.ndarray
        Unstructured numpy array
    """
    if names is None:
        names = list(live_points.dtype.names)
    return rfn.structured_to_unstructured(live_points[names], copy=copy)


def parameters_to_live_point(parameters, names, non_sampling_parameters=True):
    """
    Take a list or array of parameters for a single live point
    and converts them to a live point.

    Returns an empty array with the correct fields if len(parameters) is zero

    Parameters
    ----------
    parameters : tuple
        Float point values for each parameter
    names : tuple
        Names for each parameter as strings
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    structured_array
        Numpy structured array with fields given by names plus logP and logL
    """
    if not len(parameters):
        return empty_structured_array(
            0, names, non_sampling_parameters=non_sampling_parameters
        )
    else:
        if non_sampling_parameters:
            parameters = [
                (*parameters, *config.livepoints.non_sampling_defaults)
            ]
        else:
            parameters = [tuple(parameters)]
        return np.array(
            parameters,
            dtype=get_dtype(
                names,
                config.livepoints.default_float_dtype,
                non_sampling_parameters=non_sampling_parameters,
            ),
        )


def numpy_array_to_live_points(array, names, non_sampling_parameters=True):
    """
    Convert a numpy array to a numpy structure array with the correct fields

    Parameters
    ----------
    array : np.ndarray
        Instance of np.ndarray to convert to a structured array
    names : tuple
        Names for each parameter as strings
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    structured_array
        Numpy structured array with fields given by names plus logP and logL
    """
    if array.size == 0:
        return empty_structured_array(
            0, names=names, non_sampling_parameters=non_sampling_parameters
        )
    if array.ndim == 1:
        array = array[np.newaxis, :]
    struct_array = empty_structured_array(
        len(array),
        names=names,
        non_sampling_parameters=non_sampling_parameters,
    )
    for i, n in enumerate(names):
        struct_array[n] = array[..., i]
    return struct_array


def dict_to_live_points(d, non_sampling_parameters=True):
    """Convert a dictionary with parameters names as keys to live points.

    Assumes all entries have the same length. Also, determines number of points
    from the first entry by checking if the value has `__len__` attribute,
    if not the dictionary is assumed to contain a single point.

    Parameters
    ----------
    d : dict
        Dictionary with parameters names as keys and values that correspond
        to one or more parameters
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    structured_array
        Numpy structured array with fields given by names plus logP and logL
    """
    a = tuple(d.values())
    if hasattr(a[0], "__len__"):
        N = len(a[0])
    else:
        N = 1
    if N == 1:
        if non_sampling_parameters:
            a = (*a, *config.livepoints.non_sampling_defaults)
        return np.array(
            [a],
            dtype=get_dtype(
                d.keys(),
                config.livepoints.default_float_dtype,
                non_sampling_parameters=non_sampling_parameters,
            ),
        )
    else:
        array = empty_structured_array(
            N,
            names=list(d.keys()),
            non_sampling_parameters=non_sampling_parameters,
        )
        for k, v in d.items():
            array[k] = v
        return array


def live_points_to_dict(live_points, names=None):
    """
    Convert a structured array of live points to a dictionary with
    a key per field.

    Parameters
    ----------
    live_points : structured_array
        Array of live points
    names : list of str or None
        If None all fields in the structured array are added to the dictionary
        else only those included in the list are added.

    Returns
    -------
    dict
        Dictionary of live points
    """
    if names is None:
        names = live_points.dtype.names
    return {f: live_points[f] for f in names}


def dataframe_to_live_points(df, non_sampling_parameters=True):
    """Convert and pandas dataframe to live points.

    Adds the non-sampling parameter initialised to their defaults.

    Based on this answer on Stack Exchange:
    https://stackoverflow.com/a/51280608

    Parameters
    ----------
    df : :obj:`pandas.DataFrame`
        Pandas DataFrame to convert to live points
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    structured_array
        Numpy structured array with fields given by column names plus logP and
        logL.
    """
    if non_sampling_parameters:
        extra = config.livepoints.non_sampling_defaults
    else:
        extra = tuple()
    return np.array(
        [tuple(x) + extra for x in df.values],
        dtype=get_dtype(
            list(df.dtypes.index),
            non_sampling_parameters=non_sampling_parameters,
        ),
    )


def _unstructured_view_dtype(x, names):
    """Get the dtype for an unstructured view.

    Parameters
    ----------
    x : numpy.ndarray
        Array on which to base the dtype.
    names : Iterable
        Names for the fields to include in the dtype.

    Returns
    -------
    numpy.dtype
        Dtype containing the specified fields.
    """
    return np.dtype({name: x.dtype.fields[name] for name in names})


def unstructured_view(x, names=None, dtype=None):
    """Get an unstructured view of a live points containing certain parameters.

    This is quicker than converting to a unstructured array and does not
    create a copy of the array.

    Parameters
    ----------
    x : numpy.ndarray
        Structured array.
    names : Optional[Iterable]
        Iterable of parameters to include in the view. Must be specified if
        dtype is None.
    dtype : Optional[numpy.dtype]
        Dtype for constructing the unstructured view.

    Returns
    -------
    numpy.ndarray
        View of x as an unstructured array that contains only the
        parameters in names/dtype. Shape is (x.size, # parameters).
    """
    if dtype is None:
        dtype = _unstructured_view_dtype(x, names)
    return np.ndarray(x.shape, dtype, x, 0, x.strides).view(
        (config.livepoints.default_float_dtype, len(dtype))
    )



#!/usr/bin/env python

# Example of using the importance nested sampler in nessai with Neural
# spline flows defined on the unit-hypercube
#
# This example should take around 5 minutes to run

import numpy as np
import os

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger

output = os.path.join("outdir", "nsf_unit_hypercube")
logger = setup_logger(output=output, log_level="INFO")

# Define the model. For the importance nested sampler we must define mappings
# to and from the unit hyper-cube.


class RosenbrockModel(Model):
    """A Rosenbrock likelihood."""

    def __init__(self, dims):
        self.names = [f"x_{d}" for d in range(dims)]
        self.bounds = {n: [-5.0, 5.0] for n in self.names}

    def log_prior(self, x):
        """Log-prior"""
        log_p = np.log(self.in_bounds(x), dtype=float)
        for n in self.names:
            log_p -= np.log(self.bounds[n][1] - self.bounds[n][0])
        return log_p

    def log_likelihood(self, x):
        """Log-likelihood"""
        x = self.unstructured_view(x)
        return -(
            np.sum(
                100.0 * (x[..., 1:] - x[..., :-1] ** 2.0) ** 2.0
                + (1.0 - x[..., :-1]) ** 2.0,
                axis=-1,
            )
        )

    def to_unit_hypercube(self, x):
        """Map to the unit hyper-cube"""
        x_out = x.copy()
        for n in self.names:
            x_out[n] = (x[n] - self.bounds[n][0]) / (
                self.bounds[n][1] - self.bounds[n][0]
            )
        return x_out

    def from_unit_hypercube(self, x):
        """Map from the unit hyper-cube"""
        x_out = x.copy()
        for n in self.names:
            x_out[n] = (self.bounds[n][1] - self.bounds[n][0]) * x[
                n
            ] + self.bounds[n][0]
        return x_out


# Configure the flow to be a Neural Spline Flow with a uniform latent
# distribution
flow_config = dict(
    model_config=dict(
        n_blocks=4,
        n_neurons=32,
        ftype="nsf",
        distribution="uniform",
        kwargs=dict(
            linear_transform=None,
            batch_norm_between_layers=False,
            tail_bound=1.0,
            tails=None,
            num_bins=8,
        ),
    ),
)

# Set up the FlowSampler
# We set `reparam=None` because the flow is defined on the unit-hypercube
fs = FlowSampler(
    RosenbrockModel(4),
    nlive=10000,
    output=output,
    resume=False,
    seed=1234,
    importance_nested_sampler=True,
    draw_constant=True,
    reparameterisation=None,
    threshold_kwargs={"q": 0.66},
    reset_flow=4,
    flow_config=flow_config,
)

# And go!
fs.run()



# -*- coding: utf-8 -*-
"""
Utilities related to loading files, saving files etc.
"""
import json
import os
import shutil

import numpy as np

from ..livepoint import live_points_to_dict


def is_jsonable(x):
    """Check if an object is JSON serialisable.

    Based on: https://stackoverflow.com/a/53112659

    Parameters
    ----------
    x : obj
        Object to check

    Returns
    -------
    bool
        Boolean that indicates if the object is JSON serialisable.
    """
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False


class NessaiJSONEncoder(json.JSONEncoder):
    """Class to encode numpy arrays and other non-serialisable objects.

    Based on: https://stackoverflow.com/a/57915246.

    Examples
    --------
    This class should be used in the ``cls`` argument::

        with open(filename, 'w') as wf:
             json.dump(d, wf, indent=4, cls=NessaiJSONEncoder)
    """

    def default(self, obj):
        """Method that returns a serialisable object"""
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif not is_jsonable(obj):
            return str(obj)
        else:
            return super().default(obj)


def save_to_json(d, filename, **kwargs):
    """Save a dictionary to a JSON file.

    Kwargs are passed to :code:`json.dump`. Uses :code:`NessaiJSONEncoder` by
    default.

    Parameters
    ----------
    d : dict
        Dictionary to save.
    filename : str
        Filename (with the extension) to save the dictionary to. Should include
        the complete path.
    kwargs : Any
        Keyword arguments passed to :code:`json.dump`.
    """
    default_kwargs = dict(
        indent=4,
        cls=NessaiJSONEncoder,
    )
    default_kwargs.update(kwargs)
    with open(filename, "w") as fp:
        json.dump(d, fp, **default_kwargs)


def safe_file_dump(data, filename, module, save_existing=False):
    """Safely dump data to a .pickle file.

    See Bilby for the original implementation.

    Parameters
    ----------
    data :
        Data to dump.
    filename : str
        The file to dump to.
    module : {pickle, dill}
        The python module to use.
    save_existing : bool, optional
        If true move the existing file to <file>.old.
    """
    if save_existing:
        if os.path.exists(filename):
            old_filename = filename + ".old"
            shutil.move(filename, old_filename)

    temp_filename = filename + ".temp"
    with open(temp_filename, "wb") as file:
        module.dump(data, file)
    shutil.move(temp_filename, filename)


def save_live_points(live_points, filename):
    """Save live points to a file using JSON.

    Live points are converted to a dictionary and then saved.

    Parameters
    ----------
    live_points : ndarray
        Live points to save.
    filename : str
        File to save to.
    """
    d = live_points_to_dict(live_points)
    with open(filename, "w") as wf:
        json.dump(d, wf, indent=4, cls=NessaiJSONEncoder)


def encode_for_hdf5(value):
    """Encode a value for HDF5 file format.

    Parameters
    ----------
    value : Any
        Value to encode.

    Returns
    -------
    Any
        Encoded value.
    """
    if value is None:
        output = "__none__"
    else:
        output = value
    return output


def add_dict_to_hdf5_file(hdf5_file, path, d):
    """Save a dictionary to a HDF5 file.

    Based on :code:`recursively_save_dict_contents_to_group` in bilby.

    Parameters
    ----------
    hdf5_file : h5py.File
        HDF5 file.
    path : str
        Path added to the keys of the dictionary.
    d : dict
        The dictionary to save.
    """
    for key, value in d.items():
        if isinstance(value, dict):
            add_dict_to_hdf5_file(hdf5_file, path + key + "/", value)
        else:
            hdf5_file[path + key] = encode_for_hdf5(value)


def save_dict_to_hdf5(d, filename):
    """Save a dictionary to a HDF5 file.

    Parameters
    ----------
    d : dict
        Dictionary to save.
    filename : str
        Filename (with the extension) to save the dictionary to. Should include
        the complete path.
    """
    import h5py

    with h5py.File(filename, "w") as f:
        add_dict_to_hdf5_file(f, "/", d)



"""Tests methods for computing log_prob etc"""

from unittest.mock import MagicMock, create_autospec

from nessai.livepoint import live_points_to_array
from nessai.flowmodel.importance import ImportanceFlowModel
from nessai.proposal.importance import ImportanceFlowProposal as IFP
import numpy as np
from scipy.special import logsumexp
import pytest


@pytest.fixture()
def ifp(ifp):
    ifp.flow = create_autospec(ImportanceFlowModel)
    return ifp


def test_log_prior(ifp):
    x = np.random.randn(10, 2)
    np.testing.assert_array_equal(IFP._log_prior(ifp, x), np.zeros(10))


def test_update_proposal_weights(ifp):
    ifp._weights = {-1: 0.5, 1: 0.5}
    weights = {-1: 1 / 3, 0: 1 / 3, 1: 1 / 3}
    IFP.update_proposal_weights(ifp, weights)
    assert ifp._weights == weights


def test_update_proposal_weights_vaild(ifp):
    ifp._weights = {-1: 0.5, 1: 0.5}
    weights = {-1: 0.33, 0: 0.33, 1: 0.33}
    with pytest.raises(RuntimeError, match="Weights must sum to 1!"):
        IFP.update_proposal_weights(ifp, weights)


def test_compute_log_Q(ifp, x_prime):
    n_flows = 3
    ifp.weights_array = np.array([0.25, 0.25, 0.25, 0.25])
    ifp.flow.n_models = n_flows
    ifp.n_proposals = n_flows + 1

    log_j = np.log(np.random.rand(len(x_prime)))

    def log_prob_all(x):
        return np.log(np.random.rand(len(x), n_flows))

    ifp.flow.log_prob_all = MagicMock(side_effect=log_prob_all)

    log_Q, log_q = IFP.compute_log_Q(ifp, x_prime, log_j=log_j)

    assert len(log_Q) == len(x_prime)
    assert log_q.shape == (len(x_prime), n_flows + 1)
    assert all(log_q[:, 0] == 0)

    expected_log_Q = logsumexp(log_q, b=ifp.weights_array, axis=1)
    np.testing.assert_array_equal(log_Q, expected_log_Q)


def test_compute_log_Q_weights_not_set(ifp, x_prime):
    """Assert an error is raised if a flow is in training mode"""
    n_flows = 3
    ifp.weights = {i - 1: v for i, v in enumerate([0.25, 0.25, 0.25, np.nan])}
    ifp.flow.n_models = n_flows
    ifp.n_proposals = n_flows + 1
    log_j = np.log(np.random.rand(len(x_prime)))
    with pytest.raises(RuntimeError, match="Some weights are not set!"):
        IFP.compute_log_Q(ifp, x_prime, log_j=log_j)


def test_compute_log_Q_flow_training(ifp, x_prime):
    """Assert an error is raised if a flow is in training mode"""
    n_flows = 3
    ifp.weights_array = np.array([0.25, 0.25, 0.25, 0.25])
    ifp.flow.n_models = n_flows
    ifp.flow.models = []
    for _ in range(n_flows):
        mock_model = MagicMock()
        mock_model.training = False
        ifp.flow.models.append(mock_model)
    ifp.flow.models[-1].training = True
    ifp.n_proposals = n_flows + 1
    log_j = np.log(np.random.rand(len(x_prime)))
    with pytest.raises(
        RuntimeError, match="One or more flows are in training mode"
    ):
        IFP.compute_log_Q(ifp, x_prime, log_j=log_j)


@pytest.mark.parametrize("p_it, q_it", [(None, None), (-1, 0), (3, 4)])
def test_kl_between_proposals(ifp, model, p_it, q_it, x):
    def rescale(x):
        x = model.to_unit_hypercube(x)
        x = live_points_to_array(x, model.names)
        return x, np.zeros(x.shape[0])

    def get_proposal_log_prob(it):
        def log_prob(x):
            if it == -1:
                return np.zeros(len(x))
            else:
                return np.log(np.random.rand(len(x)))

        return log_prob

    ifp.flow.n_models = 15
    ifp.rescale = MagicMock(side_effect=rescale)
    ifp.get_proposal_log_prob = MagicMock(side_effect=get_proposal_log_prob)

    out = IFP.compute_kl_between_proposals(ifp, x, p_it, q_it)

    assert isinstance(out, float)
    assert np.isfinite(out)


def test_update_log_q(ifp, model, x):

    n_proposals = 5
    ifp.level_count = 4

    log_q = np.log(np.random.rand(len(x), n_proposals - 1))

    def rescale(x):
        x = model.to_unit_hypercube(x)
        x = live_points_to_array(x, model.names)
        return x, np.zeros(x.shape[0])

    def get_proposal_log_prob(it):
        assert it == 4

        def log_prob(x):
            return np.log(np.random.rand(len(x)))

        return log_prob

    ifp.n_proposals = n_proposals
    ifp.flow.n_models = n_proposals - 1
    ifp.rescale = MagicMock(side_effect=rescale)
    ifp.get_proposal_log_prob = MagicMock(side_effect=get_proposal_log_prob)

    log_q_out = IFP.update_log_q(ifp, x, log_q)

    assert log_q_out.shape == (len(x), n_proposals)


def test_compute_meta_proposal_from_log_q(ifp):
    n = 100
    n_prop = 10
    log_q = np.log(np.random.rand(n, n_prop))

    poolsize = np.random.multinomial(
        n_prop,
        pvals=np.ones(n_prop) / float(n_prop),
        size=n,
    )
    weights = poolsize / np.sum(poolsize)
    ifp.weights_array = weights

    expected = logsumexp(
        log_q,
        b=weights,
        axis=1,
    )

    out = IFP.compute_meta_proposal_from_log_q(ifp, log_q)

    assert len(out) == len(log_q)
    np.testing.assert_array_equal(out, expected)


@pytest.mark.usefixtures("ins_parameters")
def test_compute_meta_proposal_samples(ifp, x, x_prime, log_j):

    ifp.level_count = 2
    ifp.weights = {-1: 0.25, 0: 0.25, 1: 0.25, 2: 0.25}

    x["logQ"] = np.nan
    x["logW"] = np.nan

    log_Q = np.log(np.random.rand(len(x)))
    log_q = np.log(np.random.rand(len(x), 10))

    ifp.rescale = MagicMock(return_value=(x_prime, log_j))
    ifp.compute_log_Q = MagicMock(return_value=(log_Q, log_q))

    log_Q_out, log_q_out = IFP.compute_meta_proposal_samples(ifp, x)

    ifp.rescale.assert_called_once_with(x)
    ifp.compute_log_Q.assert_called_once_with(x_prime, log_j=log_j)

    np.testing.assert_array_equal(log_Q_out, log_Q)
    np.testing.assert_array_equal(log_q_out, log_q)


@pytest.mark.parametrize(
    "weights", [{-1: 0.5, 0: 0.5}, {-1: 0.5, 0: 0.5, 1: np.nan}]
)
@pytest.mark.usefixtures("ins_parameters")
def test_compute_meta_proposal_samples_weights_error(ifp, x, weights):
    ifp.level_count = 1
    ifp.weights = weights
    with pytest.raises(RuntimeError, match=r"Weight\(s\) missing or not set."):
        IFP.compute_meta_proposal_samples(ifp, x)



# -*- coding: utf-8 -*-
"""
Utilities for getting distributions for use in \
        :py:class:`~nessai.proposal.flowproposal.FlowProposal`
"""
from glasflow.nflows.distributions.uniform import BoxUniform
import torch
from torch.distributions import MultivariateNormal


def get_uniform_distribution(dims, r, device="cpu"):
    """
    Return a torch distribution that is uniform in the number of dims
    specified.

    Parameters
    ----------
    dims : int
        Number of dimensions.
    r : float
        Radius to use for lower and upper bounds.
    device : str, optional
        Device on which the distribution is placed.

    Returns
    -------
    :obj:`glasflow.nflows.distributions.uniform.BoxUniform`
        Instance of BoxUniform which the lower and upper bounds set by
        the radius
    """
    r = r * torch.ones(dims, device=device)
    return BoxUniform(low=-r, high=r)


def get_multivariate_normal(dims, var=1, device="cpu"):
    """
    Return a Pytorch distribution that is normally distributed in n dims
    with a given variance.

    Parameters
    ----------
    dims : int
        Number of dimensions.
    var : float, optional
        Variance.
    device : str, optional
        Device on which the distribution is placed.

    Returns
    -------
    :obj:`nessai.flows.distributions.MultivariateNormal`
        Instance of MultivariateNormal with correct variance and dims.
    """
    loc = torch.zeros(dims).to(device).double()
    covar = var * torch.eye(dims).to(device).double()
    return MultivariateNormal(loc, covariance_matrix=covar)



# -*- coding: utf-8 -*-
"""
Functions related to computing the evidence.
"""
from abc import ABC, abstractmethod
import logging
from typing import Optional

import numpy as np
import matplotlib.pyplot as plt
from scipy.special import logsumexp

from .plot import nessai_style

logger = logging.getLogger(__name__)


def logsubexp(x, y):
    """
    Helper function to compute the exponential
    of a difference between two numbers

    Computes: ``x + np.log1p(-np.exp(y-x))``

    Parameters
    ----------
    x, y : float or array_like
        Inputs
    """
    if np.any(x < y):
        raise RuntimeError(
            "cannot take log of negative number " f"{str(x)!s} - {str(y)!s}"
        )

    return x + np.log1p(-np.exp(y - x))


def log_integrate_log_trap(log_func, log_support):
    """
    Trapezoidal integration of given log(func). Returns log of the integral.

    Parameters
    ----------
    log_func : array_like
        Log values of the function to integrate over.
    log_support : array_like
        Log prior-volumes for each value.

    Returns
    -------
    float
        Log of the result of the integral.
    """
    log_func_sum = np.logaddexp(log_func[:-1], log_func[1:]) - np.log(2)
    log_dxs = logsubexp(log_support[:-1], log_support[1:])

    return logsumexp(log_func_sum + log_dxs)


class _BaseNSIntegralState(ABC):
    """Base class for the nested sampling integral."""

    @property
    @abstractmethod
    def log_evidence(self):
        """The current log-evidence."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def log_evidence_error(self):
        """The current error on the log-evidence."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def log_posterior_weights(self):
        """The log-posterior weights."""
        raise NotImplementedError()

    @property
    def effective_n_posterior_samples(self):
        """Kish's effective sample size for the posterior weights.

        Returns
        -------
        float
            The effective sample size. Returns zero if the posterior weights
            are empty.
        """
        log_p = self.log_posterior_weights
        if not len(log_p):
            return 0
        log_p -= logsumexp(log_p)
        n = np.exp(-logsumexp(2 * log_p))
        return n


class _NSIntegralState(_BaseNSIntegralState):
    """
    Stores the state of the nested sampling integrator

    Parameters
    ----------
    nlive : int
        Number of live points
    track_gradients : bool, optional
        If true the gradient of the change in logL w.r.t logX is saved each
        time `increment` is called.
    expectation : str, {logt, t}
        Method used to compute the expectation value for the shrinkage t.
        Choose between log <t> or <log t>. Defaults to <log t>.
    """

    def __init__(self, nlive, track_gradients=True, expectation="logt"):
        self.base_nlive = nlive
        self.track_gradients = track_gradients

        if expectation.lower() not in ["t", "logt"]:
            raise ValueError(
                f"Expectation must be t or logt, got: {expectation}"
            )
        self.expectation = expectation.lower()

        # Initial state of the integral
        self.logZ = -np.inf
        self.oldZ = -np.inf
        self.logw = 0
        self.info = [0.0]
        # Initially contain all the prior volume
        self.logLs = [-np.inf]  # Likelihoods sampled
        self.log_vols = [0.0]  # Volumes enclosed by contours
        self.nlive = []
        self.gradients = [0]

    @property
    def log_evidence(self):
        """The current log-evidence."""
        return self.logZ

    @property
    def log_evidence_error(self):
        """The current error on the log-evidence."""
        return np.sqrt(self.info[-1] / self.base_nlive)

    def increment(self, logL, nlive=None):
        """
        Increment the state of the evidence integrator
        Simply uses rectangle rule for initial estimate
        """
        if logL <= self.logLs[-1]:
            logger.warning(
                "NS integrator received non-monotonic logL."
                f"{self.logLs[-1]:.5f} -> {logL:.5f}"
            )
        if nlive is None:
            nlive = self.base_nlive

        self.nlive.append(nlive)
        oldZ = self.logZ
        if self.expectation == "logt":
            # <logt> approx -1 / N
            logt = -1.0 / nlive
        else:
            # <t> = N / (N + 1)
            logt = -np.log1p(1 / nlive)
        Wt = self.logw + logL + np.log1p(-np.exp(logt))
        self.logZ = np.logaddexp(self.logZ, Wt)
        # Update information estimate
        if np.isfinite(oldZ) and np.isfinite(self.logZ) and np.isfinite(logL):
            info = (
                np.exp(Wt - self.logZ) * logL
                + np.exp(oldZ - self.logZ) * (self.info[-1] + oldZ)
                - self.logZ
            )
            self.info.append(info)

        # Update history
        self.logw += logt
        self.logLs.append(logL)
        self.log_vols.append(self.logw)
        if self.track_gradients:
            self.gradients.append(
                (self.logLs[-1] - self.logLs[-2])
                / (self.log_vols[-1] - self.log_vols[-2])
            )

    def get_logx_live_points(self, nlive: int):
        """Get the log-prior volume of the current live points.

        Computes the values assuming the run were to terminate at the current
        iteration.

        Parameters
        ----------
        nlive : int
            The number of live points.

        Returns
        -------
        numpy.ndarray
            Array of log-prior volumes of length :code:`nlive`.
        """
        nlive_per_iteration = np.arange(nlive, 0, -1, dtype=float)
        if self.expectation.lower() == "logt":
            logt = -1.0 / nlive_per_iteration
        elif self.expectation.lower() == "t":
            logt = -np.log1p(1.0 / nlive_per_iteration)
        log_x = self.logw + np.cumsum(logt)
        return log_x

    def finalise(self):
        """
        Compute the final evidence with more accurate integrator
        Call at end of sampling run to refine estimate
        """
        # Trapezoidal rule
        # Extra point represents X=0 and assume max(L) = L[-1]
        self.logZ = log_integrate_log_trap(
            np.array(self.logLs + [self.logLs[-1]]),
            np.array(self.log_vols + [np.NINF]),
        )
        return self.logZ

    @nessai_style()
    def plot(self, filename=None):
        """
        Plot the logX vs logL

        Parameters
        ----------
        filename : str, optional
            Filename name for saving the figure. If not specified the figure
            is returned.
        """
        fig = plt.figure()
        plt.plot(self.log_vols, self.logLs)
        plt.title(
            f"log Z={self.logZ:.2f} "
            f"H={self.info[-1] * np.log2(np.e):.2f} bits"
        )
        plt.grid(which="both")
        plt.xlabel("log prior-volume")
        plt.ylabel("log-likelihood")
        plt.xlim([self.log_vols[-1], self.log_vols[0]])

        if filename is not None:
            fig.savefig(filename, bbox_inches="tight")
            plt.close()
            logger.debug(f"Saved nested sampling plot as {filename}")
        else:
            return fig

    @property
    def log_posterior_weights(self):
        """Compute the log-posterior weights."""
        log_L = np.array(self.logLs + [self.logLs[-1]])
        log_vols = np.array(self.log_vols + [np.NINF])
        log_Z = log_integrate_log_trap(log_L, log_vols)
        log_w = logsubexp(log_vols[:-1], log_vols[1:])
        log_post_w = log_L[1:-1] + log_w[:-1] - log_Z
        return log_post_w


class _INSIntegralState(_BaseNSIntegralState):
    """
    Object to handle computing the evidence for importance nested sampling.
    """

    def __init__(self) -> None:
        self._n_ns = 0
        self._n_lp = 0
        self._logZ = -np.inf
        self._weights_ns = None
        self._weights_lp = None
        self._weights = None

    @property
    def logZ(self) -> float:
        """The current log-evidence."""
        return self._logZ - np.log(self._n)

    log_evidence = logZ
    """Alias for logZ"""

    @property
    def log_evidence_error(self) -> float:
        """Alias for compute_uncertainty"""
        return self.compute_uncertainty(log_evidence=True)

    @property
    def evidence(self) -> float:
        """The current evidence"""
        return np.exp(self.log_evidence)

    @property
    def evidence_error(self) -> float:
        """The evidence error"""
        return self.compute_uncertainty(log_evidence=False)

    @property
    def log_evidence_live_points(self) -> float:
        """Log-evidence in the live points."""
        if self._weights_lp is None:
            raise RuntimeError("Live points are not set")
        return logsumexp(self._weights_lp) - np.log(self._weights_lp.size)

    @property
    def log_evidence_nested_samples(self):
        """Log-evidence in the nested samples"""
        return logsumexp(self._weights_ns) - np.log(self._weights_ns.size)

    @property
    def log_posterior_weights(self) -> np.ndarray:
        """Compute the log posterior weights.

        If the live points have been specified, then weights will be computed
        for these as well.
        """
        return self._weights - self.logZ

    def update_evidence(
        self,
        nested_samples: np.ndarray,
        live_points: Optional[np.ndarray] = None,
    ) -> None:
        """Update the evidence.

        Parameters
        ----------
        nested_samples
            Array of nested samples.
        live_points
            Optional array of live points, if included the evidence will
            include both live points and nested samples. If not, the evidence
            will only include the nested samples.
        """
        self._weights_ns = nested_samples["logL"] + nested_samples["logW"]
        if live_points is not None:
            self._weights_lp = live_points["logL"] + live_points["logW"]
            self._weights = np.concatenate(
                [
                    self._weights_ns,
                    self._weights_lp,
                ]
            )
        else:
            self._weights = self._weights_ns
            self._weights_lp = None
        self._logZ = logsumexp(self._weights)
        self._n = self._weights.size

    def compute_evidence_ratio(self, ns_only: bool = False) -> float:
        """
        Compute the ratio of the evidence in the live points to the nested
        samples.

        Parameters
        ----------
        ns_only
            If True only the evidence from the nested samples is used in the
            denominator of the ratio.

        Returns
        -------
        float
            Log ratio of the evidence
        """
        if ns_only:
            return (
                self.log_evidence_live_points
                - self.log_evidence_nested_samples
            )
        else:
            return self.log_evidence_live_points - self.logZ

    def compute_uncertainty(self, log_evidence: bool = True) -> float:
        """Compute the uncertainty on the current estimate of the evidence.

        Parameters
        ----------
        log_evidence :
            If true, return the uncertainty for the log-evidence rather than
            the evidence.
        """
        n = self._n
        Z_hat = np.exp(self.logZ, dtype=np.longdouble)
        Z = np.exp(self._weights, dtype=np.longdouble)
        # Standard error sqrt(Var[Z] / n)
        u = np.sqrt(np.sum((Z - Z_hat) ** 2) / (n * (n - 1)))
        if log_evidence:
            # sigma[ln Z] = |sigma[Z] / Z|
            return float(np.abs(u / Z_hat))
        else:
            return u


def log_evidence_from_ins_samples(samples: np.ndarray) -> float:
    """Compute the log-evidence for a given set of samples from the importance
    nested sampler.

    Parameters
    ----------
    samples : numpy.ndarray
        Array of samples from the importance nested samples. Must have the logW
        field.

    Returns
    -------
    float
        The log-evidence
    """
    return logsumexp(samples["logL"] + samples["logW"]) - np.log(len(samples))



# -*- coding: utf-8 -*-
"""
Definitions of common priors in the prime space.
"""
import numpy as np


def log_uniform_prior(x, xmin=-1, xmax=1):
    """Unformalised log probability of uniform prior.

    Parameters
    ----------
    x : array_like
        Parameter to computed log-prior for
    xmin : float, optional
        Lower bound on prior
    xmax : float, optional
        Upper bound on prior
    """
    return np.log((x >= xmin) & (x <= xmax))


def log_2d_cartesian_prior(x, y, k=np.pi):
    """
    Log probability for isotropic 2d Cartesian coordinates.

    Assumes a uniform distribution of angles on [0, k] and a radial component
    drawn from a chi distribution with two degrees of freedom.

    Parameters
    ----------
    x, y : array_like
        Cartesian coordinates
    k : float
        Range over which the angles used to obtain the Cartesian coordinates
        are defined.
    """
    return -np.log(k) - ((x**2 + y**2) / 2)


def log_2d_cartesian_prior_sine(x, y, k=np.pi):
    """
    Log probability of Cartesian coordinates for a angle with a sine prior

    Assumes angles drawn for a sine distribution andand a radial component
    drawn from a chi distribution with two degrees of freedom.

    Raises a RuntimeError if the anlges were not defined on the range [0, pi].

    Parameters
    ----------
    x, y : array_like
        Cartesian coordinates
    k : float
        Must be ``np.pi``. Included for compatibility with the interface for
        angle reparameterisations.
    """
    if k != np.pi:
        raise RuntimeError("x prime prior is incompatible with k != pi")
    r = x**2 + y**2
    y[y < 0] = 0
    return np.log(y / 2) - 0.5 * np.log(r) - (r / 2)


def log_3d_cartesian_prior(x, y, z):
    """
    Log probability for 3d isotropic Cartesian coordinates.

    Assumes an isotropic distribution of angles and a radial component drawn
    from a chi distribution with three degrees of freedom.

    Parameters
    ----------
    x, y, z : array_like
        Cartesian coordinates
    """
    return -1.5 * np.log(2 * np.pi) - (x**2 + y**2 + z**2) / 2



# -*- coding: utf-8 -*-
"""
Classes for interfacing with flows. This includes training and sampling.
"""
from .base import FlowModel
from .utils import update_config

__all__ = [
    "FlowModel",
    "update_config",
]



# -*- coding: utf-8 -*-
"""
Implementation of Real Non Volume Preserving flows.
"""
import logging

import numpy as np
import torch
import torch.nn.functional as F

from glasflow.nflows.distributions import StandardNormal
from glasflow.nflows import transforms

from .base import NFlow
from .utils import create_linear_transform, create_pre_transform

logger = logging.getLogger(__name__)


class RealNVP(NFlow):
    """
    Implementation of RealNVP.

    This class modifies ``SimpleRealNVP`` from nflows to allows for a custom
    mask to be parsed as a numpy array and allows for MLP to be used
    instead of a ResNet

    > L. Dinh et al., Density estimation using Real NVP, ICLR 2017.

    Parameters
    ----------
    features : int
        Number of features (dimensions) in the data space
    hidden_features : int
        Number of neurons per layer in each neural network
    num_layers : int
        Number of coupling transformations
    num_blocks_per_layer : int
        Number of layers (or blocks for resnet) per neural network for
        each coupling transform
    mask : array_like, optional
        Custom mask to use between coupling transforms. Can either be
        a single array with the same length as the number of features or
        and two-dimensional array of shape (# features, # num_layers).
        Must use -1 and 1 to indicate no updated and updated.
    context_features : int, optional
        Number of context (conditional) parameters.
    net : {'resnet', 'mlp'}
        Type of neural network to use
    use_volume_preserving : bool, optional (False)
        Use volume preserving flows which use only addition and no scaling
    activation : function
        Activation function implemented in torch
    dropout_probability : float, optional (0.0)
        Dropout probability used in each layer of the neural network
    batch_norm_within_layers : bool, optional (False)
       Enable or disable batch norm within the neural network for each coupling
       transform
    batch_norm_between_layers : bool, optional (False)
       Enable or disable batch norm between coupling transforms
    linear_transform : {'permutation', 'lu', 'svd', None}
        Linear transform to use between coupling layers. Not recommended when
        using a custom mask.
    pre_transform : str
        Linear transform to use before the first transform.
    pre_transform_kwargs : dict
        Keyword arguments to pass to the transform class used for the pre-
        transform.
    actnorm : bool
        Include activation normalisation as described in arXiv:1807.03039.
        Batch norm between layers must be disabled if using this option.
    kwargs :
        Keyword arguments are passed to the coupling class.
    """

    def __init__(
        self,
        features,
        hidden_features,
        num_layers,
        num_blocks_per_layer,
        mask=None,
        context_features=None,
        net="resnet",
        use_volume_preserving=False,
        activation=F.relu,
        dropout_probability=0.0,
        batch_norm_within_layers=False,
        batch_norm_between_layers=False,
        linear_transform=None,
        pre_transform=None,
        pre_transform_kwargs=None,
        actnorm=False,
        distribution=None,
        **kwargs,
    ):

        if features <= 1:
            raise ValueError(
                "RealNVP requires at least 2 dimensions. "
                f"Specified dimensions: {features}."
            )

        if actnorm and batch_norm_between_layers:
            raise RuntimeError(
                "Cannot enable actnorm and batchnorm between layers "
                "simultaneously."
            )

        if use_volume_preserving:
            coupling_constructor = transforms.AdditiveCouplingTransform
        else:
            coupling_constructor = transforms.AffineCouplingTransform

        if mask is None:
            mask = torch.ones(features)
            mask[::2] = -1
        else:
            mask = np.array(mask)
            if not mask.shape[-1] == features:
                raise ValueError("Mask does not match number of features")
            if mask.ndim == 2 and not mask.shape[0] == num_layers:
                raise ValueError("Mask does not match number of layers")

            mask = torch.from_numpy(mask).type(torch.get_default_dtype())

        if mask.dim() == 1:
            mask_array = torch.empty([num_layers, features])
            for i in range(num_layers):
                mask_array[i] = mask
                mask *= -1
            mask = mask_array

        if net.lower() == "resnet":
            from glasflow.nflows.nn.nets import ResidualNet

            def create_net(in_features, out_features):
                return ResidualNet(
                    in_features,
                    out_features,
                    hidden_features=hidden_features,
                    context_features=context_features,
                    num_blocks=num_blocks_per_layer,
                    activation=activation,
                    dropout_probability=dropout_probability,
                    use_batch_norm=batch_norm_within_layers,
                )

        elif net.lower() == "mlp":
            from .nets import MLP

            if batch_norm_within_layers:
                logger.warning(
                    "Batch norm within layers not supported for "
                    "MLP, will be ignored"
                )
            if dropout_probability:
                logger.warning(
                    "Dropout not supported for MLP, " "will be ignored"
                )
            hidden_features = num_blocks_per_layer * [hidden_features]

            def create_net(in_features, out_features):
                return MLP(
                    (in_features,),
                    (out_features,),
                    hidden_features,
                    activation=activation,
                )

        else:
            raise ValueError(
                f"Unknown nn type: {net}. " "Choose from: {resnet, mlp}."
            )

        layers = []

        if pre_transform is not None:
            if pre_transform_kwargs is None:
                pre_transform_kwargs = {}
            layers.append(
                create_pre_transform(
                    pre_transform, features, **pre_transform_kwargs
                )
            )

        for i in range(num_layers):

            if actnorm:
                layers.append(
                    transforms.normalization.ActNorm(features=features)
                )

            if linear_transform is not None:
                layers.append(
                    create_linear_transform(linear_transform, features)
                )
            transform = coupling_constructor(
                mask=mask[i],
                transform_net_create_fn=create_net,
                **kwargs,
            )
            layers.append(transform)
            if batch_norm_between_layers:
                layers.append(transforms.BatchNorm(features=features))

        if distribution is None:
            distribution = StandardNormal([features])

        super().__init__(
            transform=transforms.CompositeTransform(layers),
            distribution=distribution,
        )



# -*- coding: utf-8 -*-
"""
Test the AnglePair reparmeterisation
"""
import numpy as np
import pytest
from unittest.mock import MagicMock, create_autospec

from nessai.reparameterisations import AnglePair
from nessai.livepoint import (
    get_dtype,
    numpy_array_to_live_points,
    parameters_to_live_point,
)

angle_pairs = [
    (["ra", "dec"], [[0, 2 * np.pi], [-np.pi / 2, np.pi / 2]]),
    (["dec", "ra"], [[0, 2 * np.pi], [-np.pi / 2, np.pi / 2]]),
    (["ra", "dec"], [[-np.pi, np.pi], [-np.pi / 2, np.pi / 2]]),
    (["az", "zen"], [[0, 2 * np.pi], [0, np.pi]]),
    (["zen", "az"], [[0, np.pi], [0, 2 * np.pi]]),
]


@pytest.fixture(params=angle_pairs, scope="function")
def angles(request):
    return request.param


@pytest.fixture
def reparam():
    return create_autospec(AnglePair)


@pytest.fixture(scope="function")
def assert_invertibility():
    def test_invertibility(reparam, angles, radial=None):

        n = list(angles.values())[0].size
        x = np.zeros([n], dtype=get_dtype(reparam.parameters))
        x_prime = np.zeros([n], dtype=get_dtype(reparam.prime_parameters))
        log_j = 0

        for a in reparam.angles:
            x[a] = angles[a]
        if radial is not None:
            x[reparam.radial] = radial

        assert x.size == x_prime.size

        x_re, x_prime_re, log_j_re = reparam.reparameterise(x, x_prime, log_j)

        for a in reparam.angles:
            np.testing.assert_array_equal(x[a], x_re[a])
        if radial is not None:
            np.testing.assert_array_equal(
                x[reparam.radial], x_re[reparam.radial]
            )

        x_in = np.zeros([n], dtype=get_dtype(reparam.parameters))

        x_inv, x_prime_inv, log_j_inv = reparam.inverse_reparameterise(
            x_in, x_prime_re, log_j
        )

        for a in reparam.angles:
            np.testing.assert_array_almost_equal(x[a], x_inv[a])
        if radial is not None:
            np.testing.assert_array_almost_equal(
                x[reparam.radial], x_inv[reparam.radial]
            )

        np.testing.assert_array_equal(x_prime_re, x_prime_inv)
        np.testing.assert_array_almost_equal(log_j_re, -log_j_inv)

        return True

    return test_invertibility


@pytest.mark.parametrize("parameters", ["x", ["w", "x", "y", "z"]])
def test_parameters_error(parameters):
    """
    Make sure reparameterisations fails with too many or too few parameters.
    """
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair(parameters=parameters, prior_bounds=None)
    assert "Must use a pair" in str(excinfo.value)


def test_two_angles(angles):
    """Test the reparmaterisation with just the angles"""
    parameters = angles[0]
    prior_bounds = {parameters[0]: angles[1][0], parameters[1]: angles[1][1]}

    reparam = AnglePair(parameters=parameters, prior_bounds=prior_bounds)

    if "ra" in parameters:
        assert reparam.convention == "ra-dec"

    if "az" in parameters:
        assert reparam.convention == "az-zen"

    # Make sure parameter[0] is always ra or azimuth
    assert np.ptp(reparam.prior_bounds[reparam.angles[0]]) == 2 * np.pi

    assert reparam.chi is not False
    assert hasattr(reparam.chi, "rvs")
    assert reparam.has_prime_prior is False

    m = "_".join(parameters[:2])
    assert reparam.angles == parameters[:2]
    assert reparam.radial == (m + "_radial")
    assert reparam.x == (m + "_x")
    assert reparam.y == (m + "_y")
    assert reparam.z == (m + "_z")


def test_ra_dec(assert_invertibility):
    """Test the invertibility when using RA and Dec"""
    parameters = ["ra", "dec"]
    prior_bounds = {
        parameters[0]: [0, 2 * np.pi],
        parameters[1]: [-np.pi / 2, np.pi / 2],
    }

    reparam = AnglePair(
        parameters=parameters,
        prior_bounds=prior_bounds,
        convention="ra-dec",
        prior="isotropic",
    )

    n = 100
    angles = {
        "ra": np.random.uniform(*prior_bounds["ra"], n),
        "dec": np.arcsin(np.random.uniform(-1, 1, n)),
    }
    assert assert_invertibility(reparam, angles)


def test_azimuth_zenith(assert_invertibility):
    """Test the inverbility when using azimuth and zenith"""
    parameters = ["az", "zen"]
    prior_bounds = {parameters[0]: [0, 2 * np.pi], parameters[1]: [0, np.pi]}

    reparam = AnglePair(
        parameters=parameters,
        prior_bounds=prior_bounds,
        convention="az-zen",
        prior="isotropic",
    )

    n = 100
    angles = {
        "az": np.random.uniform(*prior_bounds["az"], n),
        "zen": np.arccos(np.random.uniform(-1, 1, n)),
    }
    assert reparam.parameters[:2] == list(angles.keys())
    assert assert_invertibility(reparam, angles)


def test_w_radial(assert_invertibility):
    """Test the reparameterisation with a radial parameter"""
    parameters = ["r", "ra", "dec"]
    prior_bounds = {
        parameters[0]: [0, 5],
        parameters[1]: [0, 2 * np.pi],
        parameters[2]: [-np.pi / 2, np.pi / 2],
    }

    reparam = AnglePair(parameters=parameters, prior_bounds=prior_bounds)

    assert reparam.parameters == ["ra", "dec", "r"]
    assert reparam.angles == ["ra", "dec"]
    assert reparam.chi is False

    n = 100
    angles = {
        "ra": np.random.uniform(*prior_bounds["ra"], n),
        "dec": np.arcsin(np.random.uniform(-1, 1, n)),
    }

    radial = np.random.uniform(*prior_bounds["r"], n)

    assert assert_invertibility(reparam, angles, radial=radial)


@pytest.mark.parametrize(
    "convention, input, expected",
    [
        ["az-zen", (1, 0, 0), (0, np.pi / 2, 1)],
        ["ra-dec", (1, 0, 0), (0, 0, 1)],
        ["az-zen", (-1, 0, 0), (np.pi, np.pi / 2, 1)],
        ["ra-dec", (-1, 0, 0), (np.pi, 0, 1)],
        ["az-zen", (0, 1, 0), (np.pi / 2, np.pi / 2, 1)],
        ["ra-dec", (0, 1, 0), (np.pi / 2, 0, 1)],
        ["az-zen", (0, -1, 0), (3 * np.pi / 2, np.pi / 2, 1)],
        ["ra-dec", (0, -1, 0), (3 * np.pi / 2, 0, 1)],
        ["az-zen", (0, 0, 1), (0, 0, 1)],
        ["ra-dec", (0, 0, 1), (0, np.pi / 2, 1)],
        ["az-zen", (0, 0, -1), (0, np.pi, 1)],
        ["ra-dec", (0, 0, -1), (0, -np.pi / 2, 1)],
        ["az-zen", (0, 0, 0), (0, 0, 0)],
        ["ra-dec", (0, 0, 0), (0, 0, 0)],
        ["az-zen", (1, 1, np.sqrt(2)), (np.pi / 4, np.pi / 4, 2)],
        ["ra-dec", (1, 1, np.sqrt(2)), (np.pi / 4, np.pi / 4, 2)],
        ["az-zen", (-1, -1, -np.sqrt(2)), (5 * np.pi / 4, 3 * np.pi / 4, 2)],
        ["ra-dec", (-1, -1, -np.sqrt(2)), (5 * np.pi / 4, -np.pi / 4, 2)],
    ],
)
def test_specific_points_x_prime_to_x_0_2pi(convention, input, expected):
    """Test specific points on a sphere.

    Order is (x, y, z) to (ra/az, dec/zen, r) using [0, 2pi] for ra/az.

    Test:
    - (+/-1, 0, 0) and all permutations
    - (0, 0, 0)
    - (+/-1, +-1, sqrt(2)) a point exactly 2 away from the origin
    """
    parameters = ["a", "b"]
    if convention == "ra-dec":
        prior_bounds = {"a": [0, 2 * np.pi], "b": [-np.pi / 2, np.pi / 2]}
    else:
        prior_bounds = {"a": [0, 2 * np.pi], "b": [0, np.pi]}
    reparam = AnglePair(
        parameters=parameters,
        prior_bounds=prior_bounds,
        convention=convention,
    )

    x_prime = parameters_to_live_point(input, reparam.prime_parameters)
    x = parameters_to_live_point([0, 0, 0], reparam.parameters)
    log_j = 0

    out, _, _ = reparam.inverse_reparameterise(x, x_prime, log_j)

    np.testing.assert_equal(out[reparam.parameters[0]], expected[0])
    np.testing.assert_equal(out[reparam.parameters[1]], expected[1])
    np.testing.assert_equal(out[reparam.parameters[2]], expected[2])


@pytest.mark.parametrize(
    "convention, input, expected",
    [
        ["az-zen", (1, 0, 0), (0, np.pi / 2, 1)],
        ["ra-dec", (1, 0, 0), (0, 0, 1)],
        ["az-zen", (-1, 0, 0), (np.pi, np.pi / 2, 1)],
        ["ra-dec", (-1, 0, 0), (np.pi, 0, 1)],
        ["az-zen", (0, 1, 0), (np.pi / 2, np.pi / 2, 1)],
        ["ra-dec", (0, 1, 0), (np.pi / 2, 0, 1)],
        ["az-zen", (0, -1, 0), (-np.pi / 2, np.pi / 2, 1)],
        ["ra-dec", (0, -1, 0), (-np.pi / 2, 0, 1)],
        ["az-zen", (0, 0, 1), (0, 0, 1)],
        ["ra-dec", (0, 0, 1), (0, np.pi / 2, 1)],
        ["az-zen", (0, 0, -1), (0, np.pi, 1)],
        ["ra-dec", (0, 0, -1), (0, -np.pi / 2, 1)],
        ["az-zen", (0, 0, 0), (0, 0, 0)],
        ["ra-dec", (0, 0, 0), (0, 0, 0)],
        ["az-zen", (1, 1, np.sqrt(2)), (np.pi / 4, np.pi / 4, 2)],
        ["ra-dec", (1, 1, np.sqrt(2)), (np.pi / 4, np.pi / 4, 2)],
        ["az-zen", (-1, -1, -np.sqrt(2)), (-3 * np.pi / 4, 3 * np.pi / 4, 2)],
        ["ra-dec", (-1, -1, -np.sqrt(2)), (-3 * np.pi / 4, -np.pi / 4, 2)],
    ],
)
def test_specific_points_x_prime_to_x_pi_pi(convention, input, expected):
    """Test specific points on a sphere.

    Order is (x, y, z) to (ra/az, dec/zen, r) using [-pi, pi] for ra/az.

    Test:
    - (+/-1, 0, 0) and all permutations
    - (0, 0, 0)
    - (+/-1, +-1, sqrt(2)) a point exactly 2 away from the origin
    """
    parameters = ["a", "b"]
    if convention == "ra-dec":
        prior_bounds = {"a": [-np.pi, np.pi], "b": [-np.pi / 2, np.pi / 2]}
    else:
        prior_bounds = {"a": [-np.pi, np.pi], "b": [0, np.pi]}
    reparam = AnglePair(
        parameters=parameters,
        prior_bounds=prior_bounds,
        convention=convention,
    )

    x_prime = parameters_to_live_point(input, reparam.prime_parameters)
    x = parameters_to_live_point([0, 0, 0], reparam.parameters)
    log_j = 0

    out, _, _ = reparam.inverse_reparameterise(x, x_prime, log_j)

    np.testing.assert_equal(out[reparam.parameters[0]], expected[0])
    np.testing.assert_equal(out[reparam.parameters[1]], expected[1])
    np.testing.assert_equal(out[reparam.parameters[2]], expected[2])


@pytest.mark.parametrize(
    "prior_bounds",
    [
        {"ra": [0, np.pi], "dec": [-np.pi / 2, np.pi / 2]},
        {"ra": [0, 2 * np.pi], "dec": [-np.pi, np.pi]},
    ],
)
def test_invalid_prior_ranges(prior_bounds):
    """Assert an error is raised the prior ranges are invalid"""
    with pytest.raises(ValueError) as excinfo:
        AnglePair(parameters=["ra", "dec"], prior_bounds=prior_bounds)
    assert "Invalid prior ranges" in str(excinfo.value)


def test_invalid_prior_bounds_az():
    """Assert an error is raised the prior bounds are invalid"""
    prior_bounds = {"az": [np.pi, 3 * np.pi], "zen": [0, np.pi]}
    with pytest.raises(ValueError) as excinfo:
        AnglePair(parameters=["az", "zen"], prior_bounds=prior_bounds)
    assert "Prior bounds for az must be" in str(excinfo.value)


def test_invalid_prior_bounds_inc():
    """Assert an error is raised the prior bounds are invalid for the \
        inclination angle.
    """
    prior_bounds = {"ra": [0.0, 2 * np.pi], "dec": [0, np.pi]}
    with pytest.raises(ValueError) as excinfo:
        AnglePair(
            parameters=["ra", "dec"],
            prior_bounds=prior_bounds,
            convention="ra-dec",
        )
    assert "Prior bounds for dec must be" in str(excinfo.value)


def test_unknown_prior():
    """Assert an unknown prior raises an error"""
    with pytest.raises(ValueError) as excinfo:
        AnglePair(
            parameters=["az", "zen"],
            prior_bounds={"az": [0.0, 2 * np.pi], "zen": [0.0, np.pi]},
            convention="az-zen",
            prior="uniform",
        )
    assert "Unknown prior: `uniform`. Choose from: ['isotropic', None]" in str(
        excinfo.value
    )


def test_unknown_convention():
    """Assert an unknown convention raises an error"""
    with pytest.raises(ValueError) as excinfo:
        AnglePair(
            parameters=["az", "zen"],
            prior_bounds={"az": [0.0, 2 * np.pi], "zen": [0.0, np.pi]},
            convention="sky",
        )
    assert (
        "Unknown convention: `sky`. Choose from: ['az-zen', 'ra-dec']"
        in str(excinfo.value)
    )


def test_no_convention():
    """Assert an error is raised if the convention cannot be determined."""
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair(
            parameters=["az", "zen"],
            prior_bounds={"az": [0.0, 2 * np.pi], "zen": [-np.pi, 0.0]},
            convention=None,
        )
    assert "Could not determine convention" in str(excinfo.value)


def test_log_prior():
    """Assert log_prior calls the log pdf of a chi distribution."""
    reparam = create_autospec(AnglePair)
    reparam.parameters = ["a", "b", "r"]
    reparam.has_prior = True
    x = parameters_to_live_point([0, 0, 2.0], reparam.parameters)
    reparam.chi = MagicMock()
    reparam.chi.logpdf = MagicMock(return_value=1.0)
    out = AnglePair.log_prior(reparam, x)
    reparam.chi.assert_not_called()
    reparam.chi.logpdf.assert_called_once_with(x["r"])
    assert out == 1.0


def test_log_prior_radial():
    """Assert log_prior raises an error if chi is False"""
    reparam = create_autospec(AnglePair)
    reparam.chi = False
    reparam.has_prior = True
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair.log_prior(reparam, 1)
    assert "log_prior is not defined" in str(excinfo.value)


def test_log_prior_has_prior():
    """Assert log_prior raises an error if has_prior is False.

    This shouldn't be possible unless the user defines `chi`.
    """
    reparam = create_autospec(AnglePair)
    reparam.chi = True
    reparam.has_prior = False
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair.log_prior(reparam, 1)
    assert "log_prior is not defined" in str(excinfo.value)


def test_x_prime_log_prior():
    """Test the x prime prior"""
    reparam = create_autospec(AnglePair)
    reparam.prime_parameters = ["x", "y", "z"]
    reparam.has_prime_prior = True
    x = parameters_to_live_point([1, 1, 1], reparam.prime_parameters)
    # Should be -1.5 * log(2pi) - 1.5
    expected = -1.5 * (1 + np.log(2 * np.pi))
    out = AnglePair.x_prime_log_prior(reparam, x)
    np.testing.assert_equal(out, expected)


def test_prime_prior_error():
    """Assert an error is raised if calling the prime prior and \
        has prime prior is false.
    """
    reparam = create_autospec(AnglePair)
    reparam.has_prime_prior = False
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair.x_prime_log_prior(reparam, 1)
    assert "x prime prior is not defined" in str(excinfo.value)


def test_reparameterise_negative_radius(reparam):
    """Assert an error is radius if the radius is negative."""
    x = numpy_array_to_live_points(
        np.array([[1.0, 1.0, -1.0]]), ["theta", "phi", "radius"]
    )
    x_prime = x.copy()
    log_j = 0.0
    reparam.radial = "radius"
    reparam.chi = None
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair.reparameterise(reparam, x, x_prime, log_j)
    assert "Radius cannot be negative" in str(excinfo.value)



"""Tests for the proposals and meta-proposal"""

import datetime
from unittest.mock import MagicMock, patch

from nessai.samplers.importancesampler import (
    ImportanceNestedSampler as INS,
    OrderedSamples,
)
from nessai.proposal.importance import ImportanceFlowProposal
from nessai.utils.testing import assert_structured_arrays_equal
import numpy as np
import os
import pytest


@pytest.fixture()
def ins(ins, proposal):
    ins.proposal = proposal
    return ins


def test_get_proposal(ins, model, tmp_path):
    output = str(tmp_path)
    subdir = "proposal"
    ins.model = model
    ins.output = output
    instance = object()
    with patch(
        "nessai.samplers.importancesampler.ImportanceFlowProposal",
        return_value=instance,
    ) as mock_class:
        out = INS.get_proposal(ins, subdir, test=1)
    assert out is instance
    mock_class.assert_called_once_with(
        model,
        os.path.join(output, subdir, ""),
        test=1,
    )


@pytest.mark.parametrize("weighted_kl", [False, True])
def test_add_new_proposal(ins, samples, log_q, weighted_kl):

    n = int(0.8 * len(samples))

    ins.training_samples = MagicMock(spec=OrderedSamples)
    ins.training_samples.samples = np.sort(samples, order="logL")
    ins.training_samples.log_q = log_q
    ins.log_likelihood_threshold = ins.training_samples.samples[n]["logL"]
    ins.min_samples = 2

    ins.replace_all = False
    ins.weighted_kl = weighted_kl
    ins.plot_training_data = True
    ins.training_time = datetime.timedelta(seconds=10)

    INS.add_new_proposal(ins)

    ins.proposal.train.assert_called_once()
    assert_structured_arrays_equal(
        ins.proposal.train.call_args_list[0][0][0],
        ins.training_samples.samples[n:],
    )


def test_draw_n_samples(ins, samples, log_q, history):
    expected = samples.copy()
    n = len(expected)
    ins.draw_samples_time = datetime.timedelta(seconds=10)
    ins.model.batch_evaluate_log_likelihood = MagicMock(
        return_value=samples["logL"].copy()
    )
    samples["logL"] = np.nan
    ins.proposal.draw = MagicMock(return_value=(samples, log_q))

    out = INS.draw_n_samples(ins, n)

    ins.proposal.draw.assert_called_once_with(n)

    np.testing.assert_array_equal(out[1], log_q)
    assert_structured_arrays_equal(out[0], expected)


def test_update_proposal_weights(ins):
    ins.samples_unit = np.ones(10)
    ins.sample_counts = {-1: 2, 0: 4, 1: 4}
    ins.proposal = MagicMock(spec=ImportanceFlowProposal)
    INS.update_proposal_weights(ins)
    expected_weights = {-1: 0.2, 0: 0.4, 1: 0.4}
    ins.proposal.update_proposal_weights.assert_called_once_with(
        expected_weights
    )


def test_add_new_proposal_weight(ins):
    n = 8
    n_new = 2
    sample_counts = {-1: 2, 0: 3, 1: 3}
    iteration = 2

    ins.samples_unit = np.ones(n)
    ins.sample_counts = sample_counts
    ins.proposal = MagicMock(spec=ImportanceFlowProposal)

    INS.add_new_proposal_weight(ins, iteration, n_new)

    assert ins.sample_counts[2] == 2
    expected_weights = {-1: 0.2, 0: 0.3, 1: 0.3, 2: 0.2}
    ins.proposal.update_proposal_weights.assert_called_once_with(
        expected_weights
    )


def test_add_new_proposal_weight_error(ins):
    n = 8
    n_new = 2
    sample_counts = {-1: 2, 0: 3, 1: 3, 2: 2}
    iteration = 2

    ins.samples_unit = np.ones(n)
    ins.sample_counts = sample_counts

    with pytest.raises(
        RuntimeError, match="Samples already drawn from proposal 2"
    ):
        INS.add_new_proposal_weight(ins, iteration, n_new)



# -*- coding: utf-8 -*-
"""
Test functions related to handling live points
"""
import os

import numpy as np
import pytest
from unittest.mock import MagicMock, patch

from nessai.samplers.nestedsampler import NestedSampler


def test_insert_live_point(sampler):
    """Test inserting a live point"""
    sampler.live_points = np.arange(-5, 0, 1.0).view([("logL", "f8")])
    new_point = np.array(-3.5, dtype=[("logL", "f8")])
    index = NestedSampler.insert_live_point(sampler, new_point)
    assert index == 1


def test_populate_live_points(sampler):
    """Test populating the live points"""
    samples = sampler.model.new_point(sampler.nlive)
    samples["logP"] = 0.0
    samples["logL"] = 0.0
    sampler.yield_sample = MagicMock(
        return_value=iter(zip(np.ones(sampler.nlive), samples))
    )
    NestedSampler.populate_live_points(sampler)
    assert len(sampler.live_points) == sampler.nlive


def test_populate_live_points_nans(sampler):
    """Test populating the live points with NaN values"""
    new_points = sampler.model.new_point(sampler.nlive + 1)
    new_points["logL"] = 0.0
    new_points["logL"][4] = np.nan
    new_points["logP"] = 0.0
    sampler.yield_sample = MagicMock(
        return_value=iter(zip(np.ones(sampler.nlive + 1), new_points))
    )
    NestedSampler.populate_live_points(sampler)
    assert len(sampler.live_points) == sampler.nlive
    assert not np.isnan(sampler.live_points["logL"]).any()


def test_populate_live_points_none_returned(sampler):
    """Assert that if None is returned by yield sample, it is skipped"""
    new_points = sampler.model.new_point(5)
    new_points["logL"] = 0.0
    new_points["logP"] = 0.0
    new_points = [
        None,
    ] + [*new_points]
    sampler.yield_sample = MagicMock(
        return_value=iter(zip(np.ones(len(new_points)), new_points))
    )
    sampler.nlive = 5
    NestedSampler.populate_live_points(sampler)
    np.testing.assert_array_equal(
        sampler.live_points,
        np.sort(new_points[1:], order="logL"),
    )


@pytest.mark.parametrize("rolling", [False, True])
@patch(
    "nessai.samplers.nestedsampler.compute_indices_ks_test",
    return_value=(0.1, 0.5),
)
def test_insertion_indices(mock_fn, rolling, sampler):
    """Test computing the distribution of insertion indices"""
    sampler.iteration = 100
    sampler.history = dict(rolling_p=[])
    sampler.final_p_value = None
    sampler.final_ks_statistic = None
    sampler.insertion_indices = np.random.randint(
        sampler.nlive, size=2 * sampler.nlive
    )

    NestedSampler.check_insertion_indices(sampler, rolling=rolling)

    if rolling:
        assert len(sampler.history["rolling_p"]) == 1
        np.testing.assert_array_equal(
            mock_fn.call_args_list[0][0][0],
            sampler.insertion_indices[-sampler.nlive :],
        )
        assert sampler.final_p_value is None
    else:
        mock_fn.assert_called_once_with(
            sampler.insertion_indices, sampler.nlive
        )
        assert sampler.final_p_value == 0.5
        assert sampler.final_ks_statistic == 0.1


@patch(
    "nessai.samplers.nestedsampler.compute_indices_ks_test",
    return_value=(0, None),
)
def test_insertion_indices_p_none(mock_fn, sampler):
    """Test computing the distribution of insertion indices if p is None"""
    sampler.rolling_p = []
    sampler.insertion_indices = np.random.randint(
        sampler.nlive, size=2 * sampler.nlive
    )

    NestedSampler.check_insertion_indices(sampler, rolling=True)

    assert len(sampler.rolling_p) == 0


@pytest.mark.parametrize("filename", [None, "file.txt"])
@patch("numpy.savetxt")
@patch(
    "nessai.samplers.nestedsampler.compute_indices_ks_test",
    return_value=(0.1, 0.5),
)
def test_insertion_indices_save(mock_fn, mock_save, filename, sampler):
    """Test saving the insertion indices"""
    sampler.output = os.getcwd()
    sampler.insertion_indices = np.random.randint(
        sampler.nlive, size=2 * sampler.nlive
    )

    NestedSampler.check_insertion_indices(
        sampler, rolling=False, filename=filename
    )

    if filename:
        mock_save.assert_called_once_with(
            os.path.join(os.getcwd(), "file.txt"),
            sampler.insertion_indices,
            newline="\n",
            delimiter=" ",
        )


@patch(
    "nessai.samplers.nestedsampler.compute_indices_ks_test",
    return_value=(0.1, 0.01),
)
def test_insertion_indices_p_value_warning(mock_fn, sampler, caplog):
    """Assert a warning is logged if the p-value is less than 0.05"""
    sampler.insertion_indices = [1, 2, 3]
    NestedSampler.check_insertion_indices(sampler, rolling=False)
    assert "less than 0.05" in str(caplog.text)



# -*- coding: utf-8 -*-
"""
Utilities specific to the gw subpackage.
"""
from abc import ABC, abstractmethod
import logging
import numpy as np
from scipy import interpolate

logger = logging.getLogger(__name__)

try:
    from astropy import cosmology as cosmo
    import astropy.units as u
except ImportError:
    logger.debug(
        "Could not import astropy, running with reduced functionality"
    )


class DistanceConverter(ABC):
    """Base object for converting from a distance parameter to a uniform \
        parameter.

    See :py:obj:`nessai.gw.reparameterisations.DistanceReparameterisation` \
        for more details on how the distance converters are used.
    """

    has_conversion = False
    """
    Indicates if the converter class includes a conversion. This is used when
    defining the prior in the X-prime space. For example, the
    :code:`NullDistanceConverter` doesn't apply a conversion to a
    uniform parameter, so the prime prior cannot be defined.
    """
    has_jacobian = False
    """
    Indicates if the transform applied by the converter has a tractable
    jacobian.
    """

    @abstractmethod
    def to_uniform_parameter(self, d):
        """Converter to parameter that has uniform prior.

        Parameters
        ----------
        d : array_like
            Distance.

        Returns
        -------
        d, log_j : array_like
            Distance and the log Jacobian determinant
        """
        raise NotImplementedError

    @abstractmethod
    def from_uniform_parameter(self, d):
        """Convert from a parameter that has a uniform prior.

        Parameters
        ----------
        d : array_like
            Scaled distance.

        Returns
        -------
        d, log_j : array_like
            Distance and the log Jacobian determinant
        """
        raise NotImplementedError


class NullDistanceConverter(DistanceConverter):
    """Converter that applies the identity transformation.

    Used for cases where the prior on distance is not specified.
    """

    has_jacobian = True

    def __init__(self, **kwargs):
        if kwargs:
            logger.warning(f"Kwargs {kwargs} will be ignored for distance")

    def to_uniform_parameter(self, d):
        """Applies the identity transformation.

        Parameters
        ----------
        d : array_like
            Distance.

        Returns
        -------
        d, log_j : array_like
            Distance and the log Jacobian determinant, which will always be
            zero.
        """
        return d, np.zeros_like(d)

    def from_uniform_parameter(self, d):
        """Applies the identity transformation.

        Parameters
        ----------
        d : array_like
            Distance.

        Returns
        -------
        d, log_j : array_like
            Distance and the log Jacobian determinant, which will always be
            zero.
        """
        return d, np.zeros_like(d)


class PowerLawConverter(DistanceConverter):
    """Convert from a distance parameter sampled from a power law to a uniform
    parameter.

    Assumes d is proportional to :math:`d^{(\\text{power} + 1)}` following the\
         convention in Bilby.

    Parameters
    ----------
    power : float
        Power to use for the power-law.
    scale : float
        Factor used to rescale distance prior to converting to the uniform
        parameter.
    """

    has_conversion = True
    has_jacobian = True

    def __init__(self, power=None, scale=1000.0, **kwargs):
        if power is None:
            raise RuntimeError(
                "Must specify the power to use in the power-law"
            )
        self.power = power
        self.scale = scale
        self._power = self.power + 1

        if self._power == 3:
            self._f = np.cbrt
        elif self._power == 2:
            self._f = np.sqrt
        else:
            self._f = lambda x: x ** (1 / self._power)

    def _log_jacobian(self, d):
        return (
            -self._power * np.log(self.scale)
            + np.log(self._power)
            + (self._power - 1) * np.log(d)
        )

    def _log_jacobian_inv(self, d):
        return (
            np.log(self.scale)
            - np.log(self._power)
            + (1 / self._power - 1) * np.log(d)
        )

    def to_uniform_parameter(self, d):
        """Convert distance to a parameter with a uniform prior.

        Parameters
        ----------
        d : array_like
            Distance.

        Returns
        -------
        d, log_j : array_like
            Distance and the log Jacobian determinant.
        """
        return (d / self.scale) ** (self._power), self._log_jacobian(d)

    def from_uniform_parameter(self, d):
        """Convert to distance from a parameter that has a uniform prior.

        Parameters
        ----------
        d : array_like
            scaled distance.

        Returns
        -------
        d, log_j : array_like
            Distance and the log Jacobian determinant.
        """
        return self.scale * self._f(d), self._log_jacobian_inv(d)


class ComovingDistanceConverter(DistanceConverter):
    """
    Object to convert luminosity distance with a prior that is uniform
    in co-moving volume to a parameter with a uniform prior.

    The uniform parameter is a scaled version of the co-moving distance cubed.
    This transformation does not have a tractable Jacobian determinant and
    therefore returns zero.

    Parameters
    ----------
    d_min, d_max : float
        Minimum and maximum distances.
    units : str, optional
        Units used for the distance, must be compatible with astropy units.
    cosmology : str, optional
        Cosmology used for conversion, must be compatible with astropy.
        Default is Planck15.
    scale : float, optional
        Rescaling applied to distance after converting to co-moving distance.
    pad : float, optional
        Padding used for min and max of interpolation range:
        min = (1 - pad) * d_min and max = (1 + pad) * d_max
    n_interp : int, optional
        Length of vector used for generating the look up table. For a range of
        [100, 5000] 200 seems to the minimum for the conversion to be
        invertible up to 6 decimal places. The recommended setting is at 500.
    """

    has_conversion = True
    has_jacobian = False

    def __init__(
        self,
        d_min=None,
        d_max=None,
        units="Mpc",
        cosmology="Planck15",
        scale=1000.0,
        pad=0.05,
        n_interp=500,
    ):
        self.units = u.Unit(units)
        try:
            self.cosmology = getattr(cosmo, cosmology)
            logger.info(f"Using cosmology: {cosmology}")
        except AttributeError:
            raise RuntimeError(
                f"Could not get specified cosmology ({cosmology}) from "
                "`astropy.cosmology`. See astropy documentation for details."
            )
        self.scale = np.float64(scale)
        self.pad = pad
        self.n_interp = n_interp

        self.dl_min = (1 - self.pad) * d_min
        self.dl_max = (1 + self.pad) * d_max

        logger.debug(f"Min and max distances: [{self.dl_min}, {self.dl_max}]")

        self.dc_min = self.cosmology.comoving_distance(
            cosmo.z_at_value(
                self.cosmology.luminosity_distance, self.dl_min * self.units
            )
        ).value
        self.dc_max = self.cosmology.comoving_distance(
            cosmo.z_at_value(
                self.cosmology.luminosity_distance, self.dl_max * self.units
            )
        ).value

        logger.debug("Making distance look up table")

        dc_array = np.linspace(self.dc_min, self.dc_max, self.n_interp)
        dl_array = self.cosmology.luminosity_distance(
            [
                cosmo.z_at_value(
                    self.cosmology.comoving_distance, d * self.units
                )
                for d in dc_array
            ]
        ).value

        self.interp_dc2dl = interpolate.splrep(dc_array, dl_array)
        self.interp_dl2dc = interpolate.splrep(dl_array, dc_array)

    def to_uniform_parameter(self, d):
        """Convert luminosity distance to a parameter with a uniform prior.

        Parameters
        ----------
        d : array_like
            Distance.

        Returns
        -------
        d, log_j : array_like
            Distance and the log Jacobian determinant, which will always be
            zero.
        """
        return (
            (interpolate.splev(d, self.interp_dl2dc, ext=3) / self.scale)
            ** 3.0,
            np.zeros_like(d),
        )

    def from_uniform_parameter(self, d):
        """Convert from a uniform parameter to luminosity distance.

        Parameters
        ----------
        d : array_like
            Scaled distance.

        Returns
        -------
        d, log_j : array_like
            Distance and the log Jacobian determinant, which will always be
            zero.
        """
        return (
            interpolate.splev(
                self.scale * np.cbrt(d), self.interp_dc2dl, ext=3
            ),
            np.zeros_like(d),
        )


def get_distance_converter(prior):
    """Get a distance converter from a type of prior.

    If the prior is unknown :py:obj:`nessai.gw.utils.NullDistanceConverter` \
         is returned which has the identity rescaling.

    Parameters
    ----------
    prior : str, {'uniform-comoving-volume', 'power-law'}
        The prior that is being used for the distance parameter.

    Returns
    -------
    :obj:`nessai.gw.utils.DistanceConverter`
        The corresponding distance converter.
    """
    if prior == "uniform-comoving-volume":
        return ComovingDistanceConverter
    if prior == "power-law":
        return PowerLawConverter
    else:
        logger.info(f"Prior {prior} is not known for distance")
        return NullDistanceConverter



# -*- coding: utf-8 -*-
"""
Utilities to make interfacing with bilby easier.
"""
from warnings import warn

from .settings import get_all_kwargs, get_run_kwargs_list  # noqa

msg = (
    "`nessai.utils.bilbyutils` is deprecated and will be removed in a future "
    "release. Use `nessai.utils.settings` instead."
)
warn(msg, FutureWarning)



# -*- coding: utf-8 -*-
"""
Utilities for manipulating python structures such as lists and dictionaries.
"""
import numpy as np


def replace_in_list(target_list, targets, replacements):
    """
    Replace (in place) an entry in a list with a given element.

    Parameters
    ----------
    target_list : list
        List to update
    targets : list
        List of items to update
    replacements : list
        List of replacement items
    """
    if not isinstance(targets, list):
        targets = [targets]
    if not isinstance(replacements, list):
        replacements = [replacements]

    if not len(targets) == len(replacements):
        raise RuntimeError("Targets and replacements are different lengths!")

    if not all([t in target_list for t in targets]):
        raise ValueError(f"Targets {targets} not in list: {target_list}")

    for t, r in zip(targets, replacements):
        i = target_list.index(t)
        target_list[i] = r


def get_subset_arrays(indices, *args):
    """Return a subset of a set of arrays.

    Assumes all arrays are the same length.

    Parameters
    ----------
    indices : array
        Array of indices or boolean array of same length as input arrays
    args : arrays
        Set of arrays to index.

    Returns
    -------
    tuple
        A tuple contain the corresponding array for each input array. The order
        is preserved.
    """
    return tuple(a[indices] for a in args)


def isfinite_struct(x, names=None):
    """Check for +/- infinity and NaNs in a structured array.

    Returns a boolean per entry not per field (name).

    Parameters
    ----------
    x : np.ndarray
        Structured array
    names : list[str]
        Names of the fields to include. If not specified, the names from the
        dtype of the structured array are used.

    Returns
    -------
    np.ndarray
        Array of booleans indicating if each entry in the array is finite
        (True) or not (False).
    """
    if names is None:
        names = x.dtype.names
    return np.all([np.isfinite(x[n]) for n in names], axis=0)


def array_split_chunksize(x, chunksize):
    """Split an array into multiple sub-arrays of a specified chunksize.

    Parameters
    ----------
    x : numpy.ndarray
        Input array.
    chunksize : int
        Chunksize into which to split the array.

    Returns
    -------
    list
        List of numpy arrays each with a maximum length given by the chunksize.
    """
    if chunksize < 1:
        raise ValueError("chunksize must be greater than 1")
    return np.array_split(x, range(chunksize, len(x), chunksize))


def get_inverse_indices(n, indices):
    """Return the indices that are not in input array given a size n"""
    if indices.max() >= n:
        raise ValueError("Indices contain values that are out of range for n")
    inv = np.arange(n, dtype=int)
    return inv[~np.in1d(inv, indices)]



# -*- coding: utf-8 -*-
"""
Test the null reparameterisation
"""
import numpy as np

from nessai.reparameterisations import NullReparameterisation
from nessai.livepoint import empty_structured_array, numpy_array_to_live_points
from nessai.utils.testing import (
    assert_structured_arrays_equal as assert_array_equal,
)


def test_invertiblity():
    """Ensure the reparameterisation does not change the values"""
    reparam = NullReparameterisation(parameters=["x", "y"])
    n = 100
    x = numpy_array_to_live_points(np.random.rand(n, 2), reparam.parameters)
    x_prime = empty_structured_array(n, names=reparam.prime_parameters)
    log_j = 0

    x_re, x_prime_re, log_j_re = reparam.reparameterise(x, x_prime, log_j)

    assert_array_equal(x, x_re)
    assert_array_equal(x, x_prime_re)

    x_in = empty_structured_array(n, names=reparam.parameters)
    x_inv, x_prime_inv, log_j_inv = reparam.inverse_reparameterise(
        x_in, x_prime_re, log_j
    )

    assert_array_equal(x, x_inv)
    assert_array_equal(x, x_prime_inv)
    assert log_j == log_j_re == log_j_inv



# -*- coding: utf-8 -*-
"""
Tests for torch utils.
"""
from unittest.mock import patch

import pytest
import torch

from nessai.utils.torchutils import set_torch_default_dtype


@pytest.mark.parametrize(
    "dtype, expected",
    [
        ("float32", torch.float32),
        ("float64", torch.float64),
        (torch.float32, torch.float32),
        (torch.float64, torch.float64),
    ],
)
def test_setting_default_dtype(dtype, expected):
    """Assert `set_default_dtype` is called correctly"""
    with patch("torch.set_default_dtype") as mock:
        out = set_torch_default_dtype(dtype)
    mock.assert_called_once_with(expected)
    assert out is expected


def test_setting_default_none():
    """Assert the default dtype is used when dtype is None"""
    with patch("torch.get_default_dtype", return_value=torch.float32) as mock:
        out = set_torch_default_dtype(None)
    mock.assert_called_once()
    assert out is torch.float32


def test_setting_default_dtype_invalid_input():
    """Assert an error is raised if the input is invalid"""
    with pytest.raises(ValueError) as excinfo:
        set_torch_default_dtype("not_a_dtype")
    assert "Unknown torch dtype: not_a_dtype" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""Tests related to plots in FlowProposal"""
import os
import numpy as np
import pytest
import torch
from unittest.mock import MagicMock, patch

from nessai.proposal import FlowProposal
from nessai.livepoint import numpy_array_to_live_points


@pytest.mark.parametrize("plot", [False, "all"])
def test_training_plots(proposal, tmpdir, plot):
    """Make sure traings plots are correctly produced"""
    proposal._plot_training = plot
    output = tmpdir.mkdir("test")

    names = ["x", "y"]
    prime_names = ["x_prime", "y_prime"]
    z = np.random.randn(10, 2)
    x = np.random.randn(10, 2)
    x_prime = x / 2
    proposal.training_data = numpy_array_to_live_points(x, names)
    proposal.training_data_prime = numpy_array_to_live_points(
        x_prime, prime_names
    )
    x_gen = numpy_array_to_live_points(x, names)
    x_prime_gen = numpy_array_to_live_points(x_prime, prime_names)

    # LogL will be populated before plotting
    for array in [
        proposal.training_data,
        proposal.training_data_prime,
        x_gen,
        x_prime_gen,
    ]:
        array["logL"] = 0.0

    proposal.dims = 2
    proposal.rescale_parameters = True
    proposal.parameters_to_rescale = names
    proposal.rescaled_names = prime_names

    proposal.forward_pass = MagicMock(return_value=(z, None))
    proposal.backward_pass = MagicMock(return_value=(x_prime_gen, np.ones(10)))
    proposal.inverse_rescale = MagicMock(return_value=(x_gen, np.ones(10)))
    proposal.check_prior_bounds = lambda *args: args
    proposal.model = MagicMock()
    proposal.model.names = names

    FlowProposal._plot_training_data(proposal, output)

    assert os.path.exists(os.path.join(output, "x_samples.png")) is bool(plot)
    assert os.path.exists(os.path.join(output, "x_generated.png")) is bool(
        plot
    )
    assert os.path.exists(os.path.join(output, "x_prime_samples.png")) is bool(
        plot
    )
    assert os.path.exists(
        os.path.join(output, "x_prime_generated.png")
    ) is bool(plot)


def test_plot_pool_all(proposal):
    """Test for the plots that show the pool of samples"""
    proposal.output = "test"
    proposal._plot_pool = "all"
    proposal.populated_count = 0
    x = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    with patch("nessai.proposal.flowproposal.plot_live_points") as plot:
        FlowProposal.plot_pool(proposal, x)
    plot.assert_called_once_with(
        x, c="logL", filename=os.path.join("test", "pool_0.png")
    )


@pytest.mark.parametrize("alt_dist", [False, True])
def test_plot_pool_1d(proposal, tmpdir, alt_dist):
    """Test `plot_pool` when plotting is not 'all'.

    Test cases when there is an alternative base distribution is used and
    when one is not used.
    """
    output = tmpdir.mkdir("test_plot_pool_1d")
    proposal.output = output
    proposal._plot_pool = True
    proposal.populated_count = 0

    z = np.random.randn(10, 2)
    x = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    log_q = np.random.randn(10)
    x["logP"] = np.random.randn(10)
    training_data = numpy_array_to_live_points(
        np.random.randn(10, 2), ["x", "y"]
    )
    training_data["logP"] = np.random.randn(10)
    training_data["logP"] = np.random.randn(10)
    proposal.training_data = training_data
    log_p = torch.arange(10)

    proposal.forward_pass = MagicMock(return_value=(z, log_q))
    proposal.flow = MagicMock()
    proposal.flow.device = "cpu"
    if alt_dist:
        proposal.alt_dist = MagicMock()
        proposal.alt_dist.log_prob = MagicMock(return_value=log_p)
    else:
        proposal.flow.model.base_distribution_log_prob = MagicMock(
            return_value=log_p
        )
        proposal.alt_dist = None
    with patch("nessai.proposal.flowproposal.plot_1d_comparison") as plot:
        FlowProposal.plot_pool(proposal, x)

    plot.assert_called_once_with(
        training_data,
        x,
        labels=["live points", "pool"],
        filename=os.path.join(output, "pool_0.png"),
    )
    assert os.path.exists(os.path.join(output, "pool_0_log_q.png"))

    if alt_dist:
        proposal.alt_dist.log_prob.assert_called_once()
    else:
        proposal.flow.model.base_distribution_log_prob.assert_called_once()
